from ._contracts import Contract, ChildContract, ParentContract, CONTRACT_CLASSES, CONTRACT_MAP, \
    ModelContract, SourceContract, ColumnContract, MacroContract, MacroArgumentContract
from ._core import ContractContext, ContractPart



from abc import ABCMeta, abstractmethod

from dbt_common.dataclass_schema import dbtClassMixin

from dbt_contracts.contracts.generators.properties import SetDescription
from dbt_contracts.types import DescriptionT


class SetDescriptionTester[S: DescriptionT, T: dbtClassMixin](metaclass=ABCMeta):
    """Base class for testing implementations of the SetDescription property generator"""

    @abstractmethod
    def generator(self) -> SetDescription[S, T]:
        """Fixture for the property generator to test."""
        raise NotImplementedError

    @abstractmethod
    def item(self, **kwargs) -> S:
        """Fixture for the item to test."""
        raise NotImplementedError

    @staticmethod
    def test_skips_on_empty_description(generator: SetDescription[S, T], item: S) -> None:
        original_description = item.description

        generator.overwrite = True
        generator.terminator = None

        assert not generator._set_description(item, description=None)
        assert not generator._set_description(item, description="")
        assert item.description == original_description

    @staticmethod
    def test_skips_on_not_overwrite(generator: SetDescription[S, T], item: S) -> None:
        original_description = "old description"
        item.description = original_description
        description = "new description"

        generator.overwrite = False
        generator.terminator = None

        assert not generator._set_description(item, description=description)
        assert item.description == original_description

    @staticmethod
    def test_skips_on_matching_description(generator: SetDescription[S, T], item: S) -> None:
        original_description = "description line 1\ndescription line 2"
        item.description = original_description

        generator.overwrite = True
        generator.terminator = None

        assert not generator._set_description(item, description=original_description)
        assert item.description == original_description

        generator.terminator = "\n"
        original_description_line_1 = original_description.split(generator.terminator)[0]
        item.description = original_description_line_1
        assert not generator._set_description(item, description=original_description)
        assert item.description == original_description_line_1

    @staticmethod
    def test_valid_set(generator: SetDescription[S, T], item: S) -> None:
        original_description = "old description"
        item.description = original_description
        description = "new description"

        generator.overwrite = True
        generator.terminator = None

        assert generator._set_description(item, description=description)
        assert item.description == description

        item.description = original_description
        generator.terminator = "\n"
        assert generator._set_description(item, description=description + "\n")
        assert item.description == description

        item.description = original_description
        assert generator._set_description(item, description=description + "\nanother line")
        assert item.description == description



from dbt.contracts.graph.nodes import ModelNode

from dbt_contracts.contracts._core import ContractContext
from dbt_contracts.contracts.matchers import RangeMatcher
from dbt_contracts.contracts.terms._core import validate_context
from dbt_contracts.contracts.terms.node import NodeContractTerm


class HasConstraints(NodeContractTerm[ModelNode], RangeMatcher):
    """Check whether models have an appropriate number of constraints configured in their properties."""
    @validate_context
    def run(self, item: ModelNode, context: ContractContext) -> bool:
        count = len(item.constraints)
        log_message = self._match(count=count, kind="constraints")

        if log_message:
            context.add_result(name=self.name, message=log_message, item=item)
        return not log_message



import re
from collections.abc import Sequence, Collection
from typing import Annotated, Self

from pydantic import Field, BeforeValidator, model_validator

from dbt_contracts._core import BaseModelConfig
from dbt_contracts.contracts.utils import to_tuple


class RangeMatcher(BaseModelConfig):
    min_count: int = Field(
        description="The minimum count allowed.",
        ge=1,
        default=1,
        examples=[1, 2, 3],
    )
    max_count: int | None = Field(
        description="The maximum count allowed.",
        gt=0,
        default=None,
        examples=[4, 5, 6],
    )

    @model_validator(mode="after")
    def validate_max_count(self) -> Self:
        """Ensure that the maximum count is >= the minimum count."""
        if self.max_count is not None and self.max_count < self.min_count:
            raise Exception(f"Maximum count must be >= minimum count. Got {self.max_count} > {self.min_count}")
        return self

    def _match(self, count: int, kind: str) -> str | None:
        too_small = count < self.min_count
        too_large = self.max_count is not None and count > self.max_count
        if not too_small and not too_large:
            return

        quantifier = 'few' if too_small else 'many'
        expected = self.min_count if too_small else self.max_count
        return f"Too {quantifier} {kind} found: {count}. Expected: {expected}."


class StringMatcher(BaseModelConfig):
    ignore_whitespace: bool = Field(
        description="Ignore any whitespaces when comparing data type keys.",
        default=False,
        examples=[True, False],
    )
    case_insensitive: bool = Field(
        description="Ignore cases and compare data type keys only case-insensitively.",
        default=False,
        examples=[True, False],
    )
    compare_start_only: bool = Field(
        description=(
            "Match data type keys when the two values start with the same value. "
            "Ignore the rest of the data type definition in this case."
        ),
        default=False,
        examples=[True, False],
    )

    def _match(self, actual: str | None, expected: str | None) -> bool:
        if not actual or not expected:
            return not actual and not expected

        if self.ignore_whitespace:
            actual = actual.replace(" ", "")
            expected = expected.replace(" ", "")
        if self.case_insensitive:
            actual = actual.casefold()
            expected = expected.casefold()

        if self.compare_start_only:
            match = expected.startswith(actual) or actual.startswith(expected)
        else:
            match = actual == expected

        return match


class PatternMatcher(BaseModelConfig):
    include: Annotated[Sequence[str], BeforeValidator(to_tuple)] = Field(
        description="Patterns to match against for values to include",
        default=tuple(),
        examples=[r".*i\s+am\s+a\s+regex\s+pattern.*", [r"^\w+\d+\s{1,3}$", "include[_-]this"]],
    )
    exclude: Annotated[Sequence[str], BeforeValidator(to_tuple)] = Field(
        description="Patterns to match against for values to exclude",
        default=tuple(),
        examples=[r".*i\s+am\s+a\s+regex\s+pattern.*", [r"^\w+\d+\s{1,3}$", "exclude[_-]this"]],
    )
    match_all: bool = Field(
        description="When True, all given patterns must match to be considered a match for either pattern type",
        default=False,
        examples=[True, False],
    )

    def _match(self, value: str | None) -> bool:
        if not value:
            return False
        if not self.include and not self.exclude:
            return True

        if self._match_exclude(value):
            return False
        return self._match_include(value)

    def _match_values(self, values: Collection[str | None]) -> bool:
        if not self.include and not self.exclude:
            return True

        if any(map(self._match_exclude, values)):
            return False
        return any(map(self._match_include, values))

    def _match_exclude(self, value: str | None) -> bool:
        if not value:
            return True
        if not self.exclude:
            return False

        if self.match_all and all(pattern == value or re.match(pattern, value) for pattern in self.exclude):
            return True
        if any(pattern == value or re.match(pattern, value) for pattern in self.exclude):
            return True
        return False

    def _match_include(self, value: str) -> bool:
        if not value:
            return True
        if not self.include:
            return True  # include all when no include patterns are given

        if self.match_all:
            return all(pattern == value or re.match(pattern, value) for pattern in self.include)
        return any(pattern == value or re.match(pattern, value) for pattern in self.include)



from pathlib import Path
from random import sample
from unittest import mock

import pytest
import yaml
from dbt.contracts.graph.nodes import ModelNode, SourceDefinition

from dbt_contracts.properties import PropertiesIO


class TestPropertiesIO:

    @pytest.fixture
    def properties(self) -> PropertiesIO:
        """Fixture for the PropertiesIO object to test."""
        return PropertiesIO()

    def test_get_path_with_properties_path(self, model: ModelNode):
        assert model.patch_path
        with mock.patch("dbt_contracts.properties.get_absolute_project_path") as mock_func:
            assert PropertiesIO.get_path(model) == Path(model.patch_path.split("://")[1])
            mock_func.assert_not_called()

    def test_get_path_without_properties_path(self, source: SourceDefinition):
        source.properties_path = None
        with mock.patch("dbt_contracts.properties.get_absolute_project_path") as mock_func:
            assert PropertiesIO.get_path(source) == Path(source.original_file_path)
            mock_func.assert_not_called()

    def test_get_absolute_properties_path(self, source: SourceDefinition):
        with mock.patch("dbt_contracts.properties.get_absolute_project_path") as mock_func:
            assert PropertiesIO.get_path(source, to_absolute=True).is_absolute()
            mock_func.assert_called_once()

    def test_get_properties_file_for_invalid_properties_path(
            self, properties: PropertiesIO, model: ModelNode, tmp_path: Path
    ):
        with pytest.raises(KeyError):  # properties file doesn't exist
            assert not properties[model]

        model.patch_path = None
        with pytest.raises(KeyError):  # properties path is None
            assert not properties[model]

    def test_get_properties_file_for_valid_properties_path(
            self, properties: PropertiesIO, model: ModelNode, tmp_path: Path
    ):
        path = tmp_path.joinpath(model.original_file_path).with_suffix(".yml")
        model.patch_path = f"{model.package_name}://{path}"

        path.parent.mkdir(parents=True, exist_ok=True)
        expected = {"key": "value"}
        with path.open("w") as file:
            yaml.dump(expected, file)

        with mock.patch.object(PropertiesIO, "_read_file", return_value=expected) as read_properties_file:
            assert properties[model] == expected
            read_properties_file.assert_called_once_with(path)
            assert path in properties  # loaded properties are stored

            # properties are pulled from loaded properties and file is not read again
            assert properties[model] == expected
            read_properties_file.assert_called_once_with(path)

        for path in properties:
            assert path.is_absolute()

    def test_read_properties_file(self, source: SourceDefinition, tmp_path: Path):
        path = tmp_path.joinpath(source.original_file_path)
        path.parent.mkdir(parents=True, exist_ok=True)

        output = {"key1": "value", "key2": "value"}
        with path.open("w") as file:
            yaml.dump(output, file)

        assert PropertiesIO._read_file(path) == output | {
            "__start_line__": 1,
            "__start_col__": 1,
            "__end_line__": 3,
            "__end_col__": 1,
        }

    def test_save(self, model: ModelNode, tmp_path: Path):
        properties = PropertiesIO({
            tmp_path.joinpath("properties1.yml"): {"key1": "value1"},
            tmp_path.joinpath("properties2.yml"): {"key2": "value2"},
            tmp_path.joinpath("properties3.yml"): {"key3": "value3"},
            tmp_path.joinpath("properties4.yml"): {"key4": "value4"},
        })
        paths = sample(list(properties), k=2)

        properties.save(paths)
        for path, properties in properties.items():
            if path in paths:
                assert path.is_file()
                assert yaml.full_load(path.read_text()) == properties
            else:
                assert not path.is_file()



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
from datetime import datetime
from pathlib import Path

from dbt_contracts import MODULE_ROOT, PROGRAM_OWNER_NAME, PROGRAM_OWNER_USER, PROGRAM_NAME


# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = PROGRAM_NAME
# noinspection PyShadowingBuiltins
copyright = f"{datetime.now().year}, {PROGRAM_OWNER_NAME}"
author = PROGRAM_OWNER_NAME


# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

# noinspection SpellCheckingInspection
extensions = [
    "sphinx.ext.autodoc",
    "sphinx_rtd_theme",
    "sphinx_design",
    "sphinx_new_tab_link",
]
# noinspection SpellCheckingInspection
autodoc_member_order = "bysource"
# noinspection SpellCheckingInspection
autodoc_default_options = {
    "autosummary": True,
    "members": True,
    "undoc-members": False,
    "inherited-members": False,
    "special-members": False,
    "show-inheritance": True,
    # "member-order": "bysource",
}
typehints_defaults = "braces"
typehints_use_rtype = False

templates_path = ["_templates"]
exclude_patterns = ["_build",]

suppress_warnings = [
    "sphinx_autodoc_typehints.forward_reference",
    # TODO: sphinx_autodoc_typehints.forward_reference not yet supported.
    #  See https://github.com/tox-dev/sphinx-autodoc-typehints/pull/433/files
    "sphinx.ext.inheritance_diagram",
    # TODO: Suppress 'WARNING: No classes found for inheritance diagram'
]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_static_path = ["_static"]
html_title = project

html_theme = "sphinx_rtd_theme"
html_theme_options = dict(
    collapse_navigation=False,
    sticky_navigation=True,
    navigation_depth=-1,
    includehidden=True,
    titles_only=False,
)
# html_logo = "_images/dbt-contracts-logo.png"
html_css_files = [
    "style.css",
]
html_context = dict(
    display_github=True,
    github_user=PROGRAM_OWNER_USER,
    github_repo=MODULE_ROOT,
    github_version="HEAD",
    conf_py_path=f"/{Path(__file__).parent.name}/",
)


# -- OpenGraph configuration --------------------------------------------------

ogp_site_url = f"https://{PROGRAM_OWNER_USER}.github.io/{PROGRAM_NAME.lower()}/"


# -- GraphViz configuration --------------------------------------------------
# https://graphviz.org/doc/info/attrs.html

graphviz_output_format = "svg"
inheritance_graph_attrs = dict(
    rankdir="TB", size='""', fontsize=10, ratio="auto", center="true", style="solid"
)
inheritance_node_attrs = dict(shape="ellipse", fontsize=10, fontname="monospace")



"""
Handles automatic generation of contracts reference documentation from docstrings.
"""
import shutil
from collections.abc import Collection, Iterable, Callable, Mapping
from pathlib import Path
from random import choice, sample, randrange
from types import GenericAlias, UnionType
from typing import Any

import docstring_parser
import yaml
from pydantic import BaseModel
# noinspection PyProtectedMember
from pydantic.fields import FieldInfo

from dbt_contracts.contracts import Contract, ParentContract, ChildContract, ContractPart, CONTRACT_CLASSES
from dbt_contracts.contracts.generators import PropertiesGenerator

HEADER_SECTION_CHARS = ("=", "-", "^", '"')

SECTIONS: dict[str, Callable[[type[Contract]], Collection[type[ContractPart]]]] = {
    "Filters": lambda contract: contract.__supported_conditions__,
    "Validations": lambda contract: contract.__supported_terms__,
    "Generator": lambda contract: contract.__supported_generator__ if contract.__supported_generator__ else (),
}
SECTION_DESCRIPTIONS = {
    "Filters": [
        "Filters (or Conditions) for reducing the scope of the contract.",
        "You may limit the number of {kind} processed by the rules of this contract "
        "by defining one or more of the following filters."
    ],
    "Validations": [
        "Validations (or Terms) to apply to the resources of this contract.",
        "These enforce certain standards that must be followed in order for the contract to be fulfilled."
    ],
    "Generator": [
        "You may also configure a Generator for your contract.",
        "A Generator creates/updates properties files for {kind} from the attributes found on the database resource.",
    ]
}

URL_PATH = ("reference", "contracts")


class ReferencePageBuilder:

    def __init__(self, output_dir: str | Path):
        self.output_dir = Path(output_dir)
        self.lines: list[str] = []
        self.indent = " " * 3

    def add_lines(self, lines: str | Iterable[str], indent: int = 0) -> None:
        if isinstance(lines, str):
            lines = [lines]

        if indent:
            indent_str = self.indent * indent
            lines = (indent_str + line for line in lines)

        self.lines.extend(lines)

    def add_empty_lines(self, count: int = 1) -> None:
        self.add_lines([""] * count)

    def add_code_block_lines(self, lines: str | Iterable[str], indent: int = 0) -> None:
        if isinstance(lines, str):
            lines = [lines]

        indent_str = self.indent * indent
        lines = (self.indent + indent_str + line if i != 0 else indent_str + line for i, line in enumerate(lines))

        self.lines.extend(lines)

    @staticmethod
    def make_title(value: str) -> str:
        return value.replace("_", " ").replace("-", " ").capitalize()

    def add_header(self, value: str, section: int | None = None) -> None:
        if section is None:
            header_char = HEADER_SECTION_CHARS[0]
            self.add_lines(header_char * len(value))
            self.add_lines(value)
            self.add_lines(header_char * len(value))
        elif section >= len(HEADER_SECTION_CHARS):
            self.add_lines(f"**{value}**")
        else:
            header_char = HEADER_SECTION_CHARS[section]
            self.add_lines(value)
            self.add_lines(header_char * len(value))
        self.add_empty_lines()

    @staticmethod
    def _get_description(key: str, format_map: Mapping[str, Any]) -> Iterable[str]:
        return (line.format(**format_map) for line in SECTION_DESCRIPTIONS[key])

    @staticmethod
    def _get_dropdown_block(title: str, colour: str = "primary", icon: str = None) -> list[str]:
        block = [
            f".. dropdown:: {title}",
            ":animate: fade-in",
            f":color: {colour}",
        ]
        if icon:
            block.append(f":icon: {icon}")

        block.append("")
        return block

    def generate_args(self, model: type[BaseModel], name: str):
        if not model.model_fields:
            cls = next(cls for cls in model.mro() if cls.__name__.startswith("Contract"))
            kind = cls.__name__.replace("Contract", "").split("[")[0].lower()
            no_args_doc = [
                ".. note::",
                f"This {kind} does not need further configuration. "
                f"Simply define the {kind}'s name as an item in your configuration."
            ]
            self.add_code_block_lines(no_args_doc)
            self.add_empty_lines()
            return

        self.generate_schema_ref(model, name=name)
        self.generate_example_ref(model, name=name)

    def generate_schema_ref(self, model: type[BaseModel], name: str) -> None:
        schema = self.generate_schema_dict(model)
        if not (properties := schema.get("properties")):
            return

        self.add_code_block_lines(self._get_dropdown_block("Schema", icon="gear"))

        properties_block = [".. code-block:: yaml", "", *yaml.dump({name: properties}, sort_keys=False).splitlines()]
        self.add_code_block_lines(properties_block, indent=1)
        self.add_empty_lines()

        if not (defs := self.generate_schema_dict_for_defs(schema.get("$defs"))):
            return

        self.add_code_block_lines(["**defs**", ""], indent=1)
        def_block = [".. code-block:: yaml", "", *yaml.dump(defs, sort_keys=False).splitlines()]
        self.add_code_block_lines(def_block, indent=1)
        self.add_empty_lines()

    def generate_schema_dict(self, model: type[BaseModel]) -> dict[str, Any]:
        schema = model.model_json_schema()
        self._trim_schema(schema.get("properties"))
        return schema

    def generate_schema_dict_for_defs(self, defs: dict[str, Any]) -> dict[str, Any] | None:
        if not defs:
            return

        result = {}
        for name, schema_def in defs.items():
            if not (properties := schema_def.get("properties")):
                continue
            self._trim_schema(properties)
            result[name] = properties

        return result

    @staticmethod
    def _trim_schema(schema: dict[str, Any]) -> None:
        if not schema:
            return

        for value in schema.values():
            value.pop("examples", "")
            value.pop("title", "")

    def generate_example_ref_for_contract(self, contract: type[Contract]) -> None:
        example = self.generate_example_dict_for_contract(contract)

        if issubclass(contract, ParentContract):
            example = {"contracts": {contract.__config_key__: [example]}}
        else:
            parent_contract = next(cls for cls in CONTRACT_CLASSES if cls.__child_contract__ is contract)
            example = {"contracts": {parent_contract.__config_key__: [{contract.__config_key__: [example]}]}}

        self._generate_example_dropdown(example, key="Full Example")

    @classmethod
    def generate_example_dict_for_contract(cls, contract: type[Contract]) -> dict[str, Any]:
        conditions_count = randrange(
            min(len(contract.__supported_conditions__), 3) - 1, len(contract.__supported_conditions__)
        )
        conditions = sample(contract.__supported_conditions__, k=max(1, conditions_count))

        terms_count = randrange(min(len(contract.__supported_terms__), 3) - 1, len(contract.__supported_terms__))
        terms = sample(contract.__supported_terms__, k=max(1, terms_count))

        # noinspection PyProtectedMember
        example = {
            "filter": [
                {cond._name(): cls.generate_example_dict(cond)} if cls.generate_example_dict(cond) else cond._name()
                for cond in conditions
            ],
            "validations": [
                {term._name(): cls.generate_example_dict(term)} if cls.generate_example_dict(term) else term._name()
                for term in terms
            ]
        }
        if contract.__supported_generator__ is not None:
            example["generator"] = cls.generate_example_dict(contract.__supported_generator__)

        return example

    def generate_example_ref(self, model: type[BaseModel], name: str) -> None:
        example = self.generate_example_dict(model)
        if not example:
            return

        self._generate_example_dropdown({name: example}, key="Example")

        doc = docstring_parser.parse(model.__doc__)
        if doc.description and "__EXAMPLE__" in doc.description:
            self.add_code_block_lines(doc.description.strip().split("__EXAMPLE__")[1].splitlines())
            self.add_empty_lines()

        # noinspection PyUnresolvedReferences
        field_1_name, field_1 = next(iter(model.model_fields.items()))
        if field_1_name not in example:
            return

        example = self.generate_example_dict(model)[field_1_name]
        if isinstance(example, Mapping):
            return

        first_field_example_desc = (
            f"You may also define the parameters for ``{field_1_name}`` directly on the definition like below."
        )
        self.add_code_block_lines(first_field_example_desc, indent=1)
        self.add_empty_lines()

        example_block = [".. code-block:: yaml", "", *yaml.dump({name: example}, sort_keys=False).splitlines()]
        self.add_code_block_lines(example_block, indent=1)
        self.add_empty_lines()

    def _generate_example_dropdown(self, example: dict[str, Any], key: str) -> None:
        self.add_code_block_lines(self._get_dropdown_block(key, colour="info", icon="code"))

        example_block = [".. code-block:: yaml", "", *yaml.dump(example, sort_keys=False).splitlines()]
        self.add_code_block_lines(example_block, indent=1)
        self.add_empty_lines(2)

    @classmethod
    def generate_example_dict(cls, model: type[BaseModel]) -> dict[str, Any]:
        # noinspection PyUnresolvedReferences
        examples = {name: cls._generate_example_for_field(field) for name, field in model.model_fields.items()}
        return {key: val for key, val in examples.items() if val is not None}

    @classmethod
    def _generate_example_for_field(cls, field: FieldInfo) -> Any:
        if field.examples:
            return choice(field.examples)
        elif isinstance(field.annotation, (GenericAlias, UnionType)):
            field = choice([arg for arg in field.annotation.__args__ if arg is not type(None)])
            if issubclass(field, BaseModel):
                return cls.generate_example_dict(field)
        elif issubclass(field.annotation, BaseModel):
            return cls.generate_example_dict(field.annotation)

    def generate_contract_parts(
            self, parts: type[ContractPart] | Collection[type[ContractPart]], kind: str, title: str
    ) -> None:
        description = self._get_description(kind, format_map={"kind": title.lower()})

        kind = self.make_title(kind)
        self.add_header(kind, section=0)
        if description:
            self.add_lines(description)
            self.add_empty_lines()

        if not isinstance(parts, Collection):
            title = None
            parts = [parts]

        for part in parts:
            self.generate_contract_part(part, title=title)

    def generate_contract_part(self, part: type[ContractPart], title: str | None) -> None:
        # noinspection PyProtectedMember
        name = part._name() if not issubclass(part, PropertiesGenerator) else "generator"
        if title:
            self.add_header(name, section=1)

        doc = docstring_parser.parse(part.__doc__)
        if doc.description:
            self.add_lines(doc.description.strip().split("__EXAMPLE__")[0].format(kind=title.lower()))
            self.add_empty_lines()

        self.generate_args(part, name=name)

    def generate_contract_body(self, contract: type[Contract]) -> None:
        title = self.make_title(contract.__config_key__)

        for key, getter in SECTIONS.items():
            parts = getter(contract)
            if not parts:
                continue

            self.generate_contract_parts(parts=parts, kind=key, title=title)

    def generate_ref_to_child_page(self, contract: type[ChildContract], parent_title: str) -> None:
        key = contract.__config_key__
        title = self.make_title(key)
        self.add_header(title, section=0)

        link_ref = f":ref:`{title.lower()} <{key}>`"
        description = (
            f"You may also define {title.lower().rstrip('s')}s contracts as a child set of contracts "
            f"on {parent_title.lower().rstrip('s')}s. ",
            f"Refer to the {link_ref} reference for more info."
        )

        self.add_lines(description)
        self.add_empty_lines()

    def build(self, contract: type[Contract], description: str | Iterable[str] = None) -> None:
        self.lines.clear()

        key = contract.__config_key__
        title = self.make_title(key)
        self.add_lines(f".. _{key}:")
        self.add_header(title)

        self.generate_example_ref_for_contract(contract)

        if description:
            self.add_lines(description)
            self.add_empty_lines()

        self.generate_contract_body(contract=contract)

        if issubclass(contract, ParentContract):
            self.generate_ref_to_child_page(contract.__child_contract__, parent_title=title)

        self._save(contract.__config_key__)

    def _save(self, filename: str) -> None:
        output_path = self.output_dir.joinpath(filename).with_suffix(".rst")

        self.output_dir.mkdir(parents=True, exist_ok=True)
        with output_path.open("w") as file:
            file.write("\n".join(self.lines))


if __name__ == "__main__":
    reference_pages_dir = Path(__file__).parent.joinpath(*URL_PATH)
    if reference_pages_dir.is_dir():
        shutil.rmtree(reference_pages_dir)

    builder = ReferencePageBuilder(reference_pages_dir)
    for contract_cls in CONTRACT_CLASSES:
        builder.build(contract_cls)
        if issubclass(contract_cls, ParentContract):
            builder.build(contract_cls.__child_contract__)

    print("Generated references for all contracts")



from collections.abc import Sequence, Collection, Mapping
from copy import copy
from pathlib import Path
from typing import Annotated

from dbt.artifacts.resources import BaseResource
from dbt.artifacts.resources.v1.components import ParsedResource, ColumnInfo
from dbt.artifacts.resources.v1.macro import MacroArgument
from pydantic import BeforeValidator, Field, field_validator

from dbt_contracts.contracts.conditions._core import ContractCondition
from dbt_contracts.contracts.matchers import PatternMatcher
from dbt_contracts.contracts.utils import to_tuple
from dbt_contracts.types import ItemT, TagT, MetaT


class NameCondition(ContractCondition[ItemT], PatternMatcher):
    """Filter {kind} based on their names."""
    def run(self, item: (BaseResource, ColumnInfo, MacroArgument)) -> bool:
        return self._match(item.name)


class PathCondition(ContractCondition[BaseResource], PatternMatcher):
    """
    Filter {kind} based on their paths.
    Paths must match patterns which are relative to the root directory of the dbt project.

    __EXAMPLE__

    You may define the paths as a list of lists where each part is a subdirectory within the path.
    These parts will then be unified by joining them with the os-specific path separator.
    This allows for you define os-independent configuration as needed.

    .. code-block:: yaml

        include:
        - ["path", "to", "folder1"]
        - ["path", "to", "folder2"]
        - ["path", "to", "folder3"]
        exclude:
        - ["path", "to", "another", "folder1"]
        - ["path", "to", "another", "folder2"]
        - ["path", "to", "another", "folder3"]
    """
    # noinspection PyNestedDecorators
    @field_validator("include", "exclude", mode="after", check_fields=True)
    @classmethod
    def escape_backslashes_in_windows_paths[T: Sequence[str]](cls, values: T) -> T:
        """
        Replace all single backslashes with double backslashes in Windows paths.

        This is needed to ensure regex patterns are correctly interpreted and avoid any 'bad escape' errors.
        """
        return values.__class__(path.replace("\\", "\\\\") for path in values)

    # noinspection PyNestedDecorators
    @field_validator("include", "exclude", mode="before")
    @classmethod
    def unify_chunked_path_values(cls, values: str | Sequence[str] | Sequence[Sequence[str]]) -> tuple[str, ...]:
        """
        Unify all path values into a tuple of strings.
        Also merges path parts into a single path with the os-specific separator.
        """
        if isinstance(values, str):
            return (values,)
        if not isinstance(values, Sequence):
            raise Exception(f"Unrecognised path types given: {values}")

        paths = []
        for value in values:
            if isinstance(value, str):
                paths.append(value)
                continue
            if not isinstance(value, Sequence):
                raise Exception(f"Unrecognised path types given: {values}")
            paths.append(str(Path(*value)))

        return tuple(paths)

    def run(self, item: BaseResource) -> bool:
        paths = [item.original_file_path, item.path]
        if isinstance(item, ParsedResource) and item.patch_path:
            paths.append(item.patch_path.split("://")[1])

        paths = [path.replace("\\", "\\\\") for path in paths]
        return self._match_values(paths)


class TagCondition(ContractCondition[TagT]):
    """Filter {kind} based on their tags."""
    tags: Annotated[Sequence[str], BeforeValidator(to_tuple)] = Field(
        description="The tags to match on",
        default=tuple(),
        examples=["tag1", ["tag1", "tag2"]],
    )

    def run(self, item: ParsedResource | ColumnInfo) -> bool:
        return not self.tags or any(tag in self.tags for tag in item.tags)


class MetaCondition(ContractCondition[MetaT]):
    """Filter {kind} based on their meta values."""
    meta: Mapping[str, Sequence[str]] = Field(
        description="The mapping of meta keys to their allowed values",
        default_factory=dict,
        examples=[{"key1": "val1", "key2": ["val2", "val3"]}],
    )

    # noinspection PyNestedDecorators
    @field_validator("meta", mode="before")
    @classmethod
    def make_meta_values_tuple(cls, meta: Mapping[str, str | Sequence[str]]) -> dict[str, tuple[str]]:
        """Convert all meta values to tuples"""
        meta = dict(copy(meta))

        for key, val in meta.items():
            if not isinstance(val, Collection) or isinstance(val, str):
                meta[key] = (val,)
            else:
                meta[key] = tuple(val)
        # noinspection PyTypeChecker
        return meta

    def run(self, item: ParsedResource | ColumnInfo) -> bool:
        def _match(key: str) -> bool:
            values = self.meta[key]
            if not isinstance(values, Collection) or isinstance(values, str):
                values = [values]
            return key in item.meta and item.meta[key] in values

        return not self.meta or all(map(_match, self.meta))


class IsMaterializedCondition(ContractCondition[ParsedResource]):
    """Filter {kind} taking only those which are not ephemeral."""
    def run(self, item: ParsedResource) -> bool:
        return item.config.materialized != "ephemeral"



from copy import deepcopy
from pathlib import Path
from random import choice, sample

import pytest
from _pytest.fixtures import FixtureRequest
from dbt.artifacts.resources import FileHash, BaseResource
from dbt.artifacts.resources.types import NodeType
from dbt.artifacts.resources.v1.components import ColumnInfo, ParsedResource
from dbt.artifacts.resources.v1.macro import MacroArgument
from dbt.artifacts.schemas.catalog import CatalogArtifact, CatalogMetadata
from dbt.contracts.graph.manifest import Manifest
from dbt.contracts.graph.nodes import ModelNode, SourceDefinition, Macro, TestNode, GenericTestNode, CompiledNode
from dbt_common.contracts.metadata import CatalogTable, TableMetadata, ColumnMetadata
from faker import Faker

from dbt_contracts.contracts import ContractContext


@pytest.fixture(scope="session")
def faker() -> Faker:
    """Sets up and yields a basic Faker object for fake data"""
    return Faker()


@pytest.fixture(scope="session")
def project_name(faker: Faker) -> str:
    """Fixture for the name to assign to the test project."""
    return faker.word()


@pytest.fixture(scope="session")
def context(manifest: Manifest, catalog: CatalogArtifact) -> ContractContext:
    """Fixture for a ContractContext object."""
    return ContractContext(manifest=manifest, catalog=catalog)


@pytest.fixture(scope="session")
def manifest(
        models: list[ModelNode],
        sources: list[SourceDefinition],
        macros: list[Macro],
        tests: list[TestNode],
        project_name: str,
) -> Manifest:
    """Fixture for a Manifest object."""
    manifest = Manifest()
    manifest.nodes |= {model.unique_id: model for model in models}
    manifest.nodes |= {test.unique_id: test for test in tests}
    manifest.sources |= {source.unique_id: source for source in sources}
    manifest.macros |= {macro.unique_id: macro for macro in macros}
    manifest.metadata.project_name = project_name
    return manifest


@pytest.fixture(scope="session")
def catalog(models: list[ModelNode], sources: list[SourceDefinition]) -> CatalogArtifact:
    """Fixture for a CatalogArtifact object."""
    def _generate_catalog_table(node: ParsedResource | SourceDefinition) -> CatalogTable:
        metadata = TableMetadata(type="table", schema=node.schema, name=node.name, comment=node.description)
        columns = {
            column.name: ColumnMetadata(type=column.data_type, index=idx, name=column.name, comment=column.description)
            for idx, column in enumerate(node.columns.values())
        }
        return CatalogTable(metadata=metadata, columns=columns, stats={})

    return CatalogArtifact(
        metadata=CatalogMetadata(),
        nodes={model.unique_id: _generate_catalog_table(model) for model in models},
        sources={source.unique_id: _generate_catalog_table(source) for source in sources},
    )


@pytest.fixture
def simple_resource(faker: Faker, project_name: str) -> BaseResource:
    """Fixture for a simple BaseResource object."""
    path = faker.file_path(extension=choice(("yml", "yaml", "py")), absolute=False)
    return BaseResource(
        name="_".join(faker.words()),
        path=path,
        original_file_path=str(Path("models", path)),
        package_name=project_name,
        unique_id=faker.uuid4(str),
        resource_type=NodeType.Model,
    )


@pytest.fixture(params=["model", "source"])
def node(request: FixtureRequest) -> CompiledNode:
    """Fixture for a CompiledNode object."""
    return request.getfixturevalue(request.param)


@pytest.fixture
def node_column(node: CompiledNode) -> ColumnInfo:
    """Fixture for a ColumnInfo object."""
    return choice(list(node.columns.values()))


@pytest.fixture
def catalog_table(node: CompiledNode, catalog: CatalogArtifact) -> CatalogTable:
    """Fixture for the CatalogTable object matching the current node fixture."""
    if isinstance(node, SourceDefinition):
        return catalog.sources[node.unique_id]
    return catalog.nodes[node.unique_id]


@pytest.fixture
def catalog_column(catalog_table: CatalogTable, node_column: ColumnInfo, catalog: CatalogArtifact) -> ColumnMetadata:
    """Fixture for the ColumnMetadata object matching the current node column fixture."""
    return catalog_table.columns[node_column.name]


@pytest.fixture(scope="session")
def models(faker: Faker, columns: list[ColumnInfo], project_name: str) -> list[ModelNode]:
    """Fixture for the ModelNodes that can be found in the manifest."""
    def _generate() -> ModelNode:
        path = faker.file_path(extension=choice(("sql", "py")), absolute=False)
        return ModelNode(
            name="_".join(faker.words()),
            path=path,
            original_file_path=str(Path("models", path)),
            package_name=choice([project_name, faker.word()]),
            unique_id=".".join(("models", *Path(path).with_suffix("").parts)),
            resource_type=NodeType.Model,
            alias=faker.word(),
            fqn=faker.words(3),
            checksum=FileHash(name=faker.word(), checksum="".join(faker.random_letters())),
            database=faker.word(),
            schema=faker.word(),
            patch_path=f"{faker.word()}://{faker.file_path(extension=choice(("yml", "yaml")), absolute=False)}",
            tags=faker.words(),
            meta={key: faker.word() for key in faker.words()},
            columns={column.name: column for column in sample(columns, k=faker.random_int(3, 8))},
            description=faker.sentence(),
        )

    return [_generate() for _ in range(faker.random_int(20, 30))]


@pytest.fixture
def model(models: list[ModelNode], column: ColumnInfo) -> ModelNode:
    """Fixture for a single ModelNode that can be found in the manifest."""
    return deepcopy(choice(models))


@pytest.fixture(scope="session")
def sources(faker: Faker, columns: list[ColumnInfo], project_name: str) -> list[SourceDefinition]:
    """Fixture for the SourceDefinitions that can be found in the manifest."""
    def _generate() -> SourceDefinition:
        path = faker.file_path(extension=choice(("yml", "yaml")), absolute=False)
        return SourceDefinition(
            name="_".join(faker.words()),
            path=path,
            original_file_path=str(Path("models", path)),
            package_name=choice([project_name, faker.word()]),
            unique_id=".".join(("source", *Path(path).with_suffix("").parts)),
            resource_type=NodeType.Source,
            fqn=faker.words(3),
            database=faker.word(),
            schema=faker.word(),
            identifier=faker.word(),
            source_name=faker.word(),
            source_description=faker.sentence(),
            loader=choice(("", faker.word())),
            columns={column.name: column for column in sample(columns, k=faker.random_int(3, 8))},
            description=faker.sentence(),
        )

    return [_generate() for _ in range(faker.random_int(20, 30))]


@pytest.fixture
def source(sources: list[SourceDefinition]) -> SourceDefinition:
    """Fixture for a single SourceDefinition that can be found in the manifest."""
    return deepcopy(choice(sources))


@pytest.fixture(scope="session")
def columns(faker: Faker) -> list[ColumnInfo]:
    """Fixture for the ColumnInfos that can be found in the manifest."""
    def _generate():
        data_types = (None, "varchar", "int", "timestamp", "boolean")

        return ColumnInfo(
            name="_".join(faker.words()),
            data_type=choice(data_types),
            tags=faker.words(),
            meta={key: faker.word() for key in faker.words()},
            description=faker.sentence(),
        )

    return [_generate() for _ in range(faker.random_int(40, 50))]


@pytest.fixture
def column(columns: list[ColumnInfo]) -> ColumnInfo:
    """Fixture for a single ColumnInfo that can be found in the manifest."""
    return deepcopy(choice([col for col in columns if col.name not in ("col1", "col2", "col3")]))


@pytest.fixture(scope="session")
def tests(models: list[ModelNode], sources: list[SourceDefinition], project_name: str, faker: Faker) -> list[TestNode]:
    """Fixture for the TestNodes that can be found in the manifest."""
    def _generate(item: BaseResource, column: ColumnInfo = None) -> TestNode:
        path = faker.file_path(extension=choice(("yml", "yaml", "py")), absolute=False)
        test = GenericTestNode(
            name="_".join(faker.words()),
            path=path,
            original_file_path=str(Path("tests", path)),
            package_name=choice([project_name, faker.word()]),
            unique_id=".".join(("test", *Path(path).with_suffix("").parts)),
            resource_type=NodeType.Test,
            attached_node=item.unique_id,
            alias=faker.word(),
            fqn=faker.words(3),
            checksum=FileHash(name=faker.word(), checksum="".join(faker.random_letters())),
            database=faker.word(),
            schema=faker.word(),
        )
        if column is not None:
            test.column_name = column.name

        return test

    return [
        _generate(item)
        for item in models + sources for _ in range(faker.random_int(1, 5))
    ] + [
        _generate(item, column=column)
        for item in models + sources for _ in range(faker.random_int(1, 5))
        for column in item.columns.values()
    ]


@pytest.fixture(scope="session")
def macros(faker: Faker, arguments: list[MacroArgument], project_name: str) -> list[Macro]:
    """Fixture for the Macros that can be found in the manifest."""
    def _generate() -> Macro:
        path = faker.file_path(extension="sql", absolute=False)
        return Macro(
            name="_".join(faker.words()),
            path=path,
            original_file_path=str(Path("macros", path)),
            package_name=choice([project_name, faker.word()]),
            resource_type=NodeType.Macro,
            unique_id=".".join(("macro", *Path(path).with_suffix("").parts)),
            macro_sql="SELECT * FROM table",
            arguments=sample(arguments, k=faker.random_int(3, 8)),
        )

    return [_generate() for _ in range(faker.random_int(30, 40))]


@pytest.fixture
def macro(macros: list[Macro]) -> Macro:
    """Fixture for a single Macro that can be found in the manifest."""
    return deepcopy(choice(macros))


@pytest.fixture(scope="session")
def arguments(faker: Faker) -> list[MacroArgument]:
    """Fixture for the MacroArguments that can be found in the manifest."""
    def _generate() -> MacroArgument:
        return MacroArgument(
            name="_".join(faker.words()),
        )

    return [_generate() for _ in range(faker.random_int(40, 50))]


@pytest.fixture
def argument(arguments: list[MacroArgument]) -> MacroArgument:
    """Fixture for a single MacroArguments that can be found in the manifest."""
    return deepcopy(choice([arg for arg in arguments if arg.name not in ("arg1", "arg2", "arg3")]))



"""
Invoke various dbt CLI commands needed for hooks to function and return their results.
"""
import json
import os
from argparse import Namespace
from collections.abc import Mapping
from logging import Logger
from pathlib import Path
from typing import Any

from dbt.artifacts.schemas.catalog.v1.catalog import CatalogArtifact
from dbt.cli.main import dbtRunner, dbtRunnerResult
from dbt.config import RuntimeConfig
from dbt.constants import MANIFEST_FILE_NAME
from dbt.contracts.graph.manifest import Manifest
from dbt.flags import set_from_args
from dbt.task.docs.generate import CATALOG_FILENAME
from dbt_common.context import set_invocation_context

DEFAULT_GLOBAL_ARGS = (
    "--no-use-colors",
)


def get_config(args: Namespace) -> RuntimeConfig:
    """
    Get the dbt config for the current runtime.
    The runtime config can be used to extract common dbt args for the current runtime
    e.g. project_dir, profiles_dir, target_dir etc.

    :param args: The parsed CLI args.
    :return: The runtime config.
    """
    set_invocation_context(os.environ)
    set_from_args(args, {})
    return RuntimeConfig.from_args(args)


def add_default_args(*args: str, config: RuntimeConfig | Namespace | None = None) -> list[str]:
    """
    Gets the default args to give to all commands.

    :param config: The runtime config to use for default args.
    :return: The formatted CLI args.
    """
    if config is None:
        return list(args)
    if isinstance(config, Namespace):
        config = get_config(config)

    defaults = {
        "--project-dir": config.project_root,
        "--profiles-dir": config.args.profiles_dir,
        "--profile": config.profile_name,
        "--target": config.target_name,
    }

    args = list(args)
    for key, val in defaults.items():
        if key not in args:
            args.extend((key, val))
    return args


def load_artifact(filename: str, config: RuntimeConfig | Namespace) -> Mapping[str, Any] | None:
    """
    Load an artifact from the currently configured dbt target directory.

    :param filename: The filename of the artifact to load.
    :param config: The runtime config to use when trying to load the artifact from the target path.
    :return: The loaded artifact if found. None otherwise.
    """
    if isinstance(config, Namespace):
        config = get_config(config)

    target_dir = Path(config.project_target_path)
    if not target_dir.is_dir():
        return

    target_path = target_dir.joinpath(filename)
    if not target_path.is_file():
        return

    with target_path.open("r") as file:
        artifact = json.load(file)

    return artifact


def get_result(*args, runner: dbtRunner = None) -> dbtRunnerResult:
    """
    Get the result of a dbt invocation with the given `args` and `kwargs` against a given `runner`.

    :param runner: The :py:class:`dbtRunner` to invoke commands against.
        If None, creates a new runner for this invocation.
    :param args: Args to pass to the `runner`.
    :return: The result from the invocation.
    """
    if runner is None:
        runner = dbtRunner()

    result: dbtRunnerResult = runner.invoke(list(args) + list(DEFAULT_GLOBAL_ARGS))
    if not result.success:
        raise result.exception

    return result


def clean_paths(*args, runner: dbtRunner = None, config: RuntimeConfig | Namespace = None) -> None:
    """
    Clean the configured paths i.e. run the `dbt clean` command.

    :param runner: The :py:class:`dbtRunner` to invoke commands against.
        If None, creates a new runner for this invocation.
    :param config: The runtime config to use.
    :param args: Args to pass to the `runner`.
    """
    args = add_default_args(*args, config=config)
    return get_result("clean", "--no-clean-project-files-only", *args, runner=runner).result


def install_dependencies(*args, runner: dbtRunner = None, config: RuntimeConfig | Namespace = None) -> None:
    """
    Install additional dbt dependencies i.e. run the `dbt deps` command.

    :param runner: The :py:class:`dbtRunner` to invoke commands against.
        If None, creates a new runner for this invocation.
    :param config: The runtime config to use.
    :param args: Args to pass to the `runner`.
    """
    args = add_default_args(*args, config=config)
    return get_result("deps", *args, runner=runner).result


def get_manifest(
        *args,
        runner: dbtRunner = None,
        config: RuntimeConfig | Namespace = None,
        refresh: bool = False,
        logger: Logger = None,
) -> Manifest:
    """
    Generate and return the dbt manifest for a project i.e. run the `dbt parse` command.

    :param runner: The :py:class:`dbtRunner` to invoke commands against.
        If None, creates a new runner for this invocation.
    :param config: The runtime config to use.
    :param args: Args to pass to the `runner`.
    :param refresh: Ignore any stored artifacts and generate a new manifest.
    :param logger: The logger to use for logging.
    :return: The manifest.
    """
    if not refresh:
        artifact = load_artifact(MANIFEST_FILE_NAME, config=config)
        if artifact:
            return Manifest.from_dict(artifact)

    if logger is not None:
        logger.info(f"Generating manifest for {config.project_name!r}...")

    args = add_default_args(*args, config=config)
    return get_result("parse", *args, runner=runner).result


def get_catalog(
        *args,
        runner: dbtRunner = None,
        config: RuntimeConfig | Namespace = None,
        refresh: bool = False,
        logger: Logger = None,
) -> CatalogArtifact:
    """
    Generate and return the dbt catalog for a project i.e. run the `dbt docs generate` command.

    :param runner: The :py:class:`dbtRunner` to invoke commands against.
        If None, creates a new runner for this invocation.
    :param config: The runtime config to use.
    :param args: Args to pass to the `runner`.
    :param refresh: Ignore any stored artifacts and generate a new manifest.
    :param logger: The logger to use for logging.
    :return: The catalog.
    """
    if not refresh:
        artifact = load_artifact(CATALOG_FILENAME, config=config)
        if artifact:
            return CatalogArtifact.from_dict(artifact)

    if logger is not None:
        logger.info(f"Generating catalog for {config.project_name!r}...")

    args = add_default_args(*args, config=config)
    return get_result("docs", "generate", *args, runner=runner).result



from abc import ABCMeta, abstractmethod
from collections.abc import MutableSequence
from contextlib import ExitStack
from random import sample, choice
from typing import Any
from unittest import mock

import pytest
from dbt.artifacts.resources.v1.components import ColumnInfo
from dbt.artifacts.resources.v1.macro import MacroArgument
from dbt.artifacts.schemas.catalog import CatalogArtifact
from dbt.contracts.graph.manifest import Manifest
from dbt.contracts.graph.nodes import ModelNode
from dbt.contracts.graph.nodes import SourceDefinition, Macro

from dbt_contracts.contracts import CONTRACT_CLASSES, CONTRACT_MAP
from dbt_contracts.contracts import Contract, ParentContract, ChildContract, \
    ModelContract, SourceContract, ColumnContract, MacroContract, MacroArgumentContract
from dbt_contracts.contracts.conditions import properties as c_properties, source as c_source, ContractCondition
from dbt_contracts.contracts.generators import PropertiesGenerator, ParentPropertiesGenerator, ChildPropertiesGenerator
from dbt_contracts.contracts.generators.column import ColumnPropertiesGenerator
from dbt_contracts.contracts.generators.model import ModelPropertiesGenerator
from dbt_contracts.contracts.generators.source import SourcePropertiesGenerator
from dbt_contracts.contracts.terms import properties as t_properties, source as t_source, column as t_column, \
    macro as t_macro, ContractTerm, ChildContractTerm
from dbt_contracts.types import ItemT, ParentT


class ContractTester[I: ItemT, G: PropertiesGenerator](metaclass=ABCMeta):
    """Base class for testing contracts."""
    @abstractmethod
    def contract(self, manifest: Manifest, catalog: CatalogArtifact) -> Contract[I, ContractTerm, G]:
        """Fixture for creating a contract."""
        raise NotImplementedError

    @abstractmethod
    def items(self, **kwargs) -> list[I]:
        """Fixture for items to be tested."""
        raise NotImplementedError

    @abstractmethod
    def filtered_items(self, items: list[I]) -> list[I]:
        """Fixture for filtered set of items to be tested."""
        raise NotImplementedError

    @abstractmethod
    def valid_items(self, filtered_items: list[I]) -> list[I]:
        """Fixture for valid set of items to be tested."""
        raise NotImplementedError

    @staticmethod
    def test_needs_manifest(contract: Contract[I, ContractTerm, G]):
        contract.terms = [term() for term in contract.__supported_terms__ if not term.needs_manifest]
        assert not contract.needs_manifest

        contract.terms = [term() for term in contract.__supported_terms__ if term.needs_manifest]
        if not contract.terms:
            return
        assert contract.needs_manifest

    @staticmethod
    def test_needs_catalog(contract: Contract[I, ContractTerm, G]):
        contract.terms = [term() for term in contract.__supported_terms__ if not term.needs_catalog]
        assert not contract.needs_catalog

        contract.terms = [term() for term in contract.__supported_terms__ if term.needs_catalog]
        if not contract.terms:
            return
        assert contract.needs_catalog

    @staticmethod
    def test_create_context(contract: Contract[I, ContractTerm, G]):
        assert contract.manifest is not None
        assert contract.catalog is not None

        context = contract.context
        assert context.manifest == contract.manifest
        assert context.catalog == contract.catalog
        assert not context.results

    @abstractmethod
    def _items_sort_key(self, item: I) -> Any:
        raise NotImplementedError

    def test_get_items(self, contract: Contract[I, ContractTerm, G], items: list[I], filtered_items: list[I]):
        assert sorted(contract.items, key=self._items_sort_key) == sorted(items, key=self._items_sort_key)
        assert sorted(contract.filtered_items, key=self._items_sort_key) == sorted(
            filtered_items, key=self._items_sort_key
        )

    def test_validate_items(
            self, contract: Contract[I, ContractTerm, G], filtered_items: list[I], valid_items: list[I]
    ):
        result = contract.validate()
        assert sorted(result, key=self._items_sort_key) == sorted(valid_items, key=self._items_sort_key)

        if len(filtered_items) < len(valid_items):
            assert contract.context.results

    def test_validate_items_on_no_terms(self, contract: Contract[I, ContractTerm, G], filtered_items: list[I]):
        contract.terms = []
        result = contract.validate()
        assert sorted(result, key=self._items_sort_key) == sorted(filtered_items, key=self._items_sort_key)

    @staticmethod
    def test_validate_on_selected_terms(contract: Contract[I, ContractTerm, G]):
        name = choice(contract.terms).name
        mock_map = {term.name: mock.patch.object(term.__class__, "run") for term in contract.terms}

        contract.conditions.clear()
        assert list(contract.filtered_items)

        with ExitStack() as stack:
            mocks = {mock_name: stack.enter_context(mock_term) for mock_name, mock_term in mock_map.items()}

            contract.validate(terms=[name])
            for mock_name, mock_term in mocks.items():
                mock_term.assert_called() if mock_name == name else mock_term.assert_not_called()

    @staticmethod
    def test_from_dict(contract: Contract[I, ContractTerm, G]):
        conditions = sample(
            [cls._name() for cls in contract.__supported_conditions__],
            k=min(len(contract.__supported_conditions__), 2)
        )
        terms = sample(
            [cls._name() for cls in contract.__supported_terms__],
            k=min(len(contract.__supported_terms__), 3)
        )
        generator = {"exclude": "description"}
        config = {
            choice(["filter", "conditions"]): conditions,
            choice(["validations", "terms"]): terms,
            choice(["properties", "generator"]): generator,
        }

        new_contract: Contract[I, ContractTerm] = contract.__class__.from_dict(
            config, manifest=contract.manifest, catalog=contract.catalog
        )

        assert new_contract.manifest == contract.manifest
        assert new_contract.catalog == contract.catalog

        assert [condition.name for condition in new_contract.conditions] == conditions
        assert [term.name for term in new_contract.terms] == terms
        if new_contract.__supported_generator__ is not None:
            assert new_contract.generator.exclude == (generator["exclude"],)
        else:
            assert new_contract.generator is None

    @staticmethod
    def test_create_contract_part_from_dict_with_invalid(contract: Contract[I, ContractTerm, G]):
        part_map = {cond._name(): cond for cond in contract.__supported_conditions__}
        part_name = c_source.IsEnabledCondition._name()
        part_map[part_name] = c_source.IsEnabledCondition

        # noinspection PyTypeChecker
        assert contract._create_contract_part_from_dict([part_name], part_map) is None  # bad input type
        assert contract._create_contract_part_from_dict("unknown", part_map) is None  # unrecognised part

    @staticmethod
    def test_create_contract_part_from_dict_without_config(contract: Contract[I, ContractTerm, G]):
        part_map = {cond._name(): cond for cond in contract.__supported_conditions__}
        part_name = c_source.IsEnabledCondition._name()
        part_map[part_name] = c_source.IsEnabledCondition

        with mock.patch.object(c_source.IsEnabledCondition, "__new__") as mock_condition:
            contract._create_contract_part_from_dict(part_name, part_map)
            mock_condition.assert_called_once_with(c_source.IsEnabledCondition)

    @staticmethod
    def test_create_contract_part_from_dict_with_config(contract: Contract[I, ContractTerm, G]):
        part_map = {cond._name(): cond for cond in contract.__supported_conditions__}
        part_name = c_source.IsEnabledCondition._name()
        part_map[part_name] = c_source.IsEnabledCondition

        config = {part_name: {"key1": "value", "key2": "value"}}

        with mock.patch.object(c_source.IsEnabledCondition, "__new__") as mock_condition:
            contract._create_contract_part_from_dict(config, part_map)
            mock_condition.assert_called_once_with(c_source.IsEnabledCondition, **config[part_name])


class ParentContractTester[I: ItemT, P: ParentT, G: ParentPropertiesGenerator](ContractTester[P, G]):
    """Base class for testing parent contracts."""
    @abstractmethod
    def contract(self, manifest: Manifest, catalog: CatalogArtifact) -> ParentContract[I, P, G]:
        raise NotImplementedError

    @abstractmethod
    def child_conditions(self) -> MutableSequence[ContractCondition]:
        """Fixture for child conditions."""
        raise NotImplementedError

    @abstractmethod
    def child_terms(self) -> MutableSequence[ChildContractTerm]:
        """Fixture for child terms."""
        raise NotImplementedError

    @pytest.fixture
    def child_generator(self) -> ChildPropertiesGenerator | None:
        """Fixture for a child generator."""
        return

    def _items_sort_key(self, item: P) -> Any:
        return item.unique_id

    @staticmethod
    def test_contract_is_mapped(contract: ParentContract[I, P, G]):
        assert contract.__class__ in CONTRACT_CLASSES
        assert CONTRACT_MAP.get(contract.__config_key__) == contract.__class__

    @staticmethod
    def test_validate_terms(contract: ParentContract[I, P, G]):
        assert contract.validate_terms(contract.terms)
        if contract.__supported_terms__:
            assert not contract.validate_terms([])

        invalid_classes = [
            cls for cls in contract.__child_contract__.__supported_terms__
            if cls not in contract.__supported_terms__
        ]
        if not invalid_classes:
            return

        invalid_cls = choice(invalid_classes)
        assert not contract.validate_terms(list(contract.terms) + [invalid_cls()])

        with pytest.raises(Exception):
            contract.__class__(terms=list(contract.terms) + [invalid_cls()])

    @staticmethod
    def test_validate_conditions(contract: ParentContract[I, P, G]):
        assert contract.validate_conditions(contract.conditions)
        if contract.__supported_conditions__:
            assert not contract.validate_conditions([])

        invalid_classes = [
            cls for cls in contract.__child_contract__.__supported_conditions__
            if cls not in contract.__supported_conditions__
        ]
        if not invalid_classes:
            return

        invalid_cls = choice(invalid_classes)
        assert not contract.validate_conditions(list(contract.conditions) + [invalid_cls()])

        with pytest.raises(Exception):
            contract.__class__(conditions=list(contract.conditions) + [invalid_cls()])

    @staticmethod
    def test_create_child_contract(
            contract: ParentContract[I, P, G],
            child_conditions: MutableSequence[ContractCondition],
            child_terms: MutableSequence[ChildContractTerm],
            child_generator: ChildPropertiesGenerator,
    ):
        child = contract.create_child_contract(
            conditions=child_conditions, terms=child_terms, generator=child_generator
        )
        assert child.parent == contract
        assert child.conditions == child_conditions
        assert child.terms == child_terms
        assert child.generator == child_generator

    @staticmethod
    def test_create_child_contract_from_dict(contract: ParentContract[I, P, G]):
        config = {
            "filter": sample(
                [cls._name() for cls in contract.__supported_conditions__],
                k=min(len(contract.__supported_conditions__), 2)
            ),
            "terms": sample(
                [cls._name() for cls in contract.__supported_terms__],
                k=min(len(contract.__supported_terms__), 3)
            ),
            contract.__child_contract__.__config_key__: [{
                "filter": sample(
                    [cls._name() for cls in contract.__child_contract__.__supported_conditions__],
                    k=min(len(contract.__child_contract__.__supported_conditions__), 2)
                ),
                "terms": sample(
                    [cls._name() for cls in contract.__child_contract__.__supported_terms__],
                    k=min(len(contract.__child_contract__.__supported_terms__), 3)
                ),
            }]
        }

        child = contract.create_child_contract_from_dict(config)[0]

        assert child.parent == contract
        assert child.manifest == contract.manifest
        assert child.catalog == contract.catalog

        child_config = config[contract.__child_contract__.__config_key__][0]
        assert [condition.name for condition in child.conditions] == child_config["filter"]
        assert [term.name for term in child.terms] == child_config["terms"]

    @staticmethod
    def test_validate_context_for_manifest(contract: ParentContract[I, P, G], items: list[I], manifest: Manifest):
        contract.terms = [term() for term in contract.__supported_terms__ if term.needs_manifest]
        context = contract.context
        context.manifest = None
        for term in contract.terms:
            with pytest.raises(Exception):
                term.run(choice(items), context=context)

        context.manifest = manifest
        for term in contract.terms:  # just check that these don't fail
            term.run(choice(items), context=context)

    @staticmethod
    def test_validate_context_for_catalog(contract: ParentContract[I, P, G], items: list[I], catalog: CatalogArtifact):
        contract.terms = [term() for term in contract.__supported_terms__ if not term.needs_catalog]
        assert not contract.needs_catalog

        contract.terms = [term() for term in contract.__supported_terms__ if term.needs_catalog]
        if not contract.terms:
            return
        assert contract.needs_catalog

        context = contract.context
        context.catalog = None
        for term in contract.terms:
            with pytest.raises(Exception):
                term.run(choice(items), context=context)

        context.catalog = catalog
        for term in contract.terms:  # just check that these don't fail
            term.run(choice(items), context=context)

    @staticmethod
    def test_generate(contract: Contract[I, ContractTerm, G], filtered_items: list[I]):
        contract.generator = None
        assert not contract.generate()

        if contract.__supported_generator__ is None:
            return

        contract.generator = contract.__supported_generator__()

        with (
                mock.patch.object(
                    contract.__class__, "filtered_items", new_callable=mock.PropertyMock, return_value=filtered_items
                ),
                mock.patch.object(contract.generator.__class__, "merge", return_value=True) as mock_merge,
                mock.patch.object(contract.generator.__class__, "update") as mock_update,
        ):
            results = contract.generate()
            assert results
            assert sum(results.values()) == len(filtered_items)

            mock_merge.assert_called()
            assert len(mock_merge.mock_calls) == len(filtered_items)
            mock_update.assert_called()
            assert len(mock_update.mock_calls) == len(filtered_items)

        with (
                mock.patch.object(contract.generator.__class__, "merge", return_value=False) as mock_merge,
                mock.patch.object(contract.generator.__class__, "update") as mock_update,
        ):
            results = contract.generate()
            assert not results

            mock_merge.assert_called()
            assert len(mock_merge.mock_calls) == len(list(contract.filtered_items))
            mock_update.assert_not_called()


class ChildContractTester[I: ItemT, P: ParentT, G: ChildPropertiesGenerator](ContractTester[I, G]):
    @abstractmethod
    def items(self, parent: ParentContract[I, P, G], **kwargs) -> list[tuple[I, P]]:
        raise NotImplementedError

    @abstractmethod
    def filtered_items(self, items: list[tuple[I, P]]) -> list[tuple[I, P]]:
        raise NotImplementedError

    @abstractmethod
    def valid_items(self, filtered_items: list[tuple[I, P]]) -> list[tuple[I, P]]:
        raise NotImplementedError

    # noinspection PyMethodOverriding
    @abstractmethod
    def contract(
            self, manifest: Manifest, catalog: CatalogArtifact, parent: ParentContract[I, P, ParentPropertiesGenerator]
    ) -> ChildContract[I, P, G]:
        raise NotImplementedError

    @abstractmethod
    def parent(self, manifest: Manifest, catalog: CatalogArtifact) -> ParentContract[I, P, ParentPropertiesGenerator]:
        """Fixture for parent contract."""
        raise NotImplementedError

    def _items_sort_key(self, item: tuple[I, P]) -> Any:
        return item[1].unique_id, item[0].name

    @staticmethod
    def test_validate_terms(contract: ChildContract[I, P, G], parent: ParentContract[I, P, ParentPropertiesGenerator]):
        assert contract.validate_terms(contract.terms)

        invalid_classes = [
            cls for cls in parent.__supported_terms__
            if cls not in contract.__supported_terms__
        ]
        if not invalid_classes:
            return

        invalid_cls = choice(invalid_classes)
        assert not contract.validate_terms(list(contract.terms) + [invalid_cls()])

        with pytest.raises(Exception):
            contract.__class__(terms=list(contract.terms) + [invalid_cls()])

    @staticmethod
    def test_validate_conditions(
            contract: ChildContract[I, P, G], parent: ParentContract[I, P, ParentPropertiesGenerator]
    ):
        assert contract.validate_conditions(contract.conditions)

        invalid_classes = [
            cls for cls in parent.__supported_conditions__
            if cls not in contract.__supported_conditions__
        ]
        if not invalid_classes:
            return

        invalid_cls = choice(invalid_classes)
        assert not contract.validate_conditions(list(contract.conditions) + [invalid_cls()])

        with pytest.raises(Exception):
            contract.__class__(conditions=list(contract.conditions) + [invalid_cls()])

    @staticmethod
    def test_validate_context_for_manifest(
            contract: ChildContract[I, P, G], items: list[tuple[I, P]], manifest: Manifest
    ):
        contract.terms = [term() for term in contract.__supported_terms__ if term.needs_manifest]
        context = contract.context
        context.manifest = None
        for term in contract.terms:
            with pytest.raises(Exception):
                item, parent = choice(items)
                term.run(item, parent=parent, context=context)

        context.manifest = manifest
        for term in contract.terms:  # just check that these don't fail
            item, parent = choice(items)
            term.run(item, parent=parent, context=context)

    @staticmethod
    def test_validate_context_for_catalog(
            contract: ChildContract[I, P, G], items: list[tuple[I, P]], catalog: CatalogArtifact
    ):
        contract.terms = [term() for term in contract.__supported_terms__ if not term.needs_catalog]
        assert not contract.needs_catalog

        contract.terms = [term() for term in contract.__supported_terms__ if term.needs_catalog]
        if not contract.terms:
            return
        assert contract.needs_catalog

        context = contract.context
        context.catalog = None
        for term in contract.terms:
            with pytest.raises(Exception):
                item, parent = choice(items)
                term.run(item, parent=parent, context=context)

        context.catalog = catalog
        for term in contract.terms:  # just check that these don't fail
            item, parent = choice(items)
            term.run(item, parent=parent, context=context)

    @staticmethod
    def test_generate(contract: ChildContract[I, P, G], filtered_items: list[tuple[I, P]]):
        contract.generator = None
        assert not contract.generate()

        if contract.__supported_generator__ is None:
            return

        contract.generator = contract.__supported_generator__()
        contract.parent.generator = contract.parent.__supported_generator__()

        with (
                mock.patch.object(
                    contract.__class__, "filtered_items", new_callable=mock.PropertyMock, return_value=filtered_items
                ),
                mock.patch.object(contract.generator.__class__, "merge", return_value=True) as mock_merge,
                mock.patch.object(contract.parent.generator.__class__, "update") as mock_update,
        ):
            results = contract.generate()
            assert results
            assert sum(results.values()) == len(filtered_items)

            mock_merge.assert_called()
            assert len(mock_merge.mock_calls) == len(filtered_items)
            mock_update.assert_called()
            assert len(mock_update.mock_calls) == len(filtered_items)

        with (
                mock.patch.object(contract.generator.__class__, "merge", return_value=False) as mock_merge,
                mock.patch.object(contract.parent.generator.__class__, "update") as mock_update,
        ):
            results = contract.generate()
            assert not results

            mock_merge.assert_called()
            assert len(mock_merge.mock_calls) == len(list(contract.filtered_items))
            mock_update.assert_not_called()


class TestModelContract(ParentContractTester[ColumnInfo, ModelNode, ModelPropertiesGenerator]):
    @pytest.fixture(scope="class")
    def items(self, models: list[ModelNode], manifest: Manifest) -> list[ModelNode]:
        return [model for model in models if model.package_name == manifest.metadata.project_name]

    @pytest.fixture(scope="class")
    def filtered_items(self, items: list[ModelNode]) -> list[ModelNode]:
        assert len(items) > 3

        items = sample(items, k=len(items) // 3)
        for item in items:
            item.name = choice(("model1", "model2"))
            item.tags.append("include")

        assert len(items) > 0
        return items

    @pytest.fixture(scope="class")
    def valid_items(self, filtered_items: list[ModelNode]) -> list[ModelNode]:
        items = sample(filtered_items, k=len(filtered_items) // 2)
        for item in items:
            item.tags.append("required_tag")

        return items

    @pytest.fixture
    def contract(self, manifest: Manifest, catalog: CatalogArtifact) -> ModelContract:
        conditions = [
            c_properties.NameCondition(include=["model1", "model2"]),
            c_properties.TagCondition(tags=["include"])
        ]
        terms = [
            t_properties.HasRequiredTags(tags="required_tag"),
        ]
        return ModelContract(manifest=manifest, catalog=catalog, conditions=conditions, terms=terms)

    @pytest.fixture(scope="class")
    def child_conditions(self) -> MutableSequence[ContractCondition]:
        return [
            c_properties.NameCondition(include=["col1", "col2"]),
            c_properties.TagCondition(tags=["valid"])
        ]

    @pytest.fixture(scope="class")
    def child_terms(self) -> MutableSequence[ContractTerm]:
        return [
            t_column.HasDataType(min_count=3),
        ]

    @pytest.fixture(scope="class")
    def child_generator(self) -> ModelPropertiesGenerator:
        return ModelPropertiesGenerator()


class TestSourceContract(ParentContractTester[ColumnInfo, SourceDefinition, SourcePropertiesGenerator]):
    @pytest.fixture(scope="class")
    def items(self, sources: list[SourceDefinition], manifest: Manifest) -> list[SourceDefinition]:
        return [source for source in sources if source.package_name == manifest.metadata.project_name]

    @pytest.fixture(scope="class")
    def filtered_items(self, items: list[SourceDefinition]) -> list[SourceDefinition]:
        assert len(items) > 3

        items = sample(items, k=len(items) // 3)
        for item in items:
            item.name = choice(("source1", "source2"))
            item.tags.append("include")

        assert len(items) > 0
        return items

    @pytest.fixture(scope="class")
    def valid_items(self, filtered_items: list[SourceDefinition]) -> list[SourceDefinition]:
        items = [item for item in filtered_items if bool(item.loader)]
        items = sample(items, k=len(items) // 2)
        for item in items:
            item.tags.append("required_tag")

        return items

    @pytest.fixture
    def contract(self, manifest: Manifest, catalog: CatalogArtifact) -> SourceContract:
        conditions = [
            c_properties.NameCondition(include=["source1", "source2"]),
            c_properties.TagCondition(tags=["include"])
        ]
        terms = [
            t_properties.HasRequiredTags(tags="required_tag"),
            t_source.HasLoader(),
        ]
        return SourceContract(manifest=manifest, catalog=catalog, conditions=conditions, terms=terms)

    @pytest.fixture(scope="class")
    def child_conditions(self) -> MutableSequence[ContractCondition]:
        return [
            c_properties.NameCondition(include=["col1", "col2"]),
            c_properties.TagCondition(tags=["valid"])
        ]

    @pytest.fixture(scope="class")
    def child_terms(self) -> MutableSequence[ContractTerm]:
        return [
            t_column.HasDataType(min_count=3),
        ]

    @pytest.fixture(scope="class")
    def child_generator(self) -> SourcePropertiesGenerator:
        return SourcePropertiesGenerator()


class TestColumnContract(ChildContractTester[ColumnInfo, ModelNode, ColumnPropertiesGenerator]):
    # noinspection PyTestUnpassedFixture
    @pytest.fixture(scope="class")
    def items(self, parent: ModelContract, **__) -> list[tuple[ColumnInfo, ModelNode]]:
        parent_items = list(parent.filtered_items)
        assert parent_items
        return [(col, item) for item in parent_items for col in item.columns.values()]

    @pytest.fixture(scope="class")
    def filtered_items(self, items: list[tuple[ColumnInfo, ModelNode]]) -> list[tuple[ColumnInfo, ModelNode]]:
        assert len(items) > 3

        columns = list({col.name: col for col, _ in items}.values())
        columns = sample(columns, k=len(columns) // 3)
        for col in columns:
            col.tags.append("valid")

        assert len(columns) > 0
        return [(col, parent) for col, parent in items if "valid" in col.tags]

    @pytest.fixture(scope="class")
    def valid_items(self, filtered_items: list[tuple[ColumnInfo, ModelNode]]) -> list[tuple[ColumnInfo, ModelNode]]:
        return [
            (col, parent) for col, parent in filtered_items if bool(col.data_type)
        ]

    @pytest.fixture
    def contract(self, manifest: Manifest, catalog: CatalogArtifact, parent: ModelContract) -> ColumnContract:
        conditions = [
            c_properties.TagCondition(tags=["valid"])
        ]
        terms = [
            t_column.HasDataType(),
        ]
        return ColumnContract(parent=parent, conditions=conditions, terms=terms)

    @pytest.fixture(scope="class")
    def parent(self, manifest: Manifest, catalog: CatalogArtifact) -> ModelContract:
        return ModelContract(manifest=manifest, catalog=catalog)


class TestMacroContract(ParentContractTester[MacroArgument, Macro, None]):
    @pytest.fixture(scope="class")
    def items(self, macros: list[Macro], manifest: Manifest) -> list[Macro]:
        return [macro for macro in macros if macro.package_name == manifest.metadata.project_name]

    @pytest.fixture(scope="class")
    def filtered_items(self, items: list[Macro]) -> list[Macro]:
        assert len(items) > 3

        items = sample(items, k=len(items) // 3)
        for item in items:
            item.name = choice(("macro1", "macro2"))

        assert len(items) > 0
        return items

    @pytest.fixture(scope="class")
    def valid_items(self, filtered_items: list[Macro]) -> list[Macro]:
        return [
            item for item in filtered_items if bool(item.description)
        ]

    @pytest.fixture
    def contract(self, manifest: Manifest, catalog: CatalogArtifact) -> MacroContract:
        conditions = [
            c_properties.NameCondition(include=["macro1", "macro2"]),
        ]
        terms = [
            t_properties.HasDescription(),
        ]
        return MacroContract(manifest=manifest, catalog=catalog, conditions=conditions, terms=terms)

    @pytest.fixture(scope="class")
    def child_conditions(self) -> MutableSequence[ContractCondition]:
        return [
            c_properties.NameCondition(include=["arg1", "arg2"]),
        ]

    @pytest.fixture(scope="class")
    def child_terms(self) -> MutableSequence[ContractTerm]:
        return [
            t_macro.HasType(),
        ]


class TestMacroArgumentContract(ChildContractTester[MacroArgument, Macro, None]):
    # noinspection PyTestUnpassedFixture
    @pytest.fixture(scope="class")
    def items(self, parent: MacroContract, **__) -> list[tuple[MacroArgument, Macro]]:
        parent_items = list(parent.filtered_items)
        assert parent_items
        return [(arg, item) for item in parent_items for arg in item.arguments]

    @pytest.fixture(scope="class")
    def filtered_items(self, items: list[tuple[MacroArgument, Macro]]) -> list[tuple[MacroArgument, Macro]]:
        assert len(items) > 3

        arguments = list({arg.name: arg for arg, _ in items}.values())
        arguments = sample(arguments, k=len(arguments) // 3)
        for arg in arguments:
            arg.name = choice(("arg1", "arg2"))

        assert len(arguments) > 0
        return [(arg, parent) for arg, parent in items if arg.name in ("arg1", "arg2")]

    @pytest.fixture(scope="class")
    def valid_items(self, filtered_items: list[tuple[MacroArgument, Macro]]) -> list[tuple[MacroArgument, Macro]]:
        return [
            (arg, parent) for arg, parent in filtered_items if bool(arg.type)
        ]

    @pytest.fixture
    def contract(self, manifest: Manifest, catalog: CatalogArtifact, parent: MacroContract) -> MacroArgumentContract:
        conditions = [
            c_properties.NameCondition(include=["arg1", "arg2"]),
        ]
        terms = [
            t_macro.HasType(),
        ]
        return MacroArgumentContract(parent=parent, conditions=conditions, terms=terms)

    @pytest.fixture(scope="class")
    def parent(self, manifest: Manifest, catalog: CatalogArtifact) -> MacroContract:
        return MacroContract(manifest=manifest, catalog=catalog)



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from typing import ClassVar

from dbt.artifacts.schemas.catalog import CatalogArtifact
from dbt.contracts.graph.manifest import Manifest

from dbt_contracts._core import BaseModelConfig
from dbt_contracts.contracts.result import Result, RESULT_PROCESSOR_MAP
from dbt_contracts.properties import PropertiesIO
from dbt_contracts.types import ItemT, ParentT


class ContractPart(BaseModelConfig, metaclass=ABCMeta):
    """
    A part of a contract.

    A ContractPart may only process the items it is given.
    It does not know how to get those items from any dbt artifacts.
    These items must be provided by a :py:class:`.Contract` object.
    """

    @property
    def name(self) -> str:
        """The name of this contract part in snake_case."""
        return self._name()

    @classmethod
    @abstractmethod
    def _name(cls) -> str:
        """The name of this contract part in snake_case."""
        raise NotImplementedError


@dataclass
class ContractContext:
    """
    Context for a contract to run within.
    Stores artifacts for the loaded DBT project and handles logging of results.
    """
    manifest: Manifest | None = None
    catalog: CatalogArtifact | None = None
    properties: ClassVar[PropertiesIO] = PropertiesIO()

    @property
    def results(self) -> list[Result]:
        """The list of stored results from term validations."""
        return self._results

    def __post_init__(self) -> None:
        self._results = []

    def add_result(self, name: str, message: str, item: ItemT, parent: ParentT = None, **kwargs) -> None:
        """
        Create and add a new :py:class:`.Result` to the current list

        :param name: The name to give to the generated result.
        :param message: The result message.
        :param item: The item that produced the result.
        :param parent: The parent of the item that produced the result if available.
        :param kwargs: Other result kwargs to pass to the result
        """
        processor = RESULT_PROCESSOR_MAP.get(type(item))
        if processor is None:
            raise Exception(f"Unexpected item to create result for: {type(item)}")

        result = processor.from_resource(
            item=item,
            parent=parent,
            properties=self.properties,
            result_name=name,
            result_level="warning",
            message=message,
            **kwargs
        )
        self.results.append(result)



import argparse
import logging
import os
from copy import deepcopy
from pathlib import Path

from dbt.cli.resolvers import default_profiles_dir, default_project_dir

from dbt_contracts import PROGRAM_NAME
from dbt_contracts.contracts import CONTRACT_CLASSES
from dbt_contracts.dbt_cli import get_config, clean_paths, install_dependencies, get_manifest, get_catalog
from dbt_contracts.runner import ContractsRunner

CORE_PARSER = argparse.ArgumentParser(
    prog=PROGRAM_NAME,
)

################################################################################
## DBT args
################################################################################
profiles_dir = CORE_PARSER.add_argument(
    "--profiles-dir",
    help="Which directory to look in for the profiles.yml file. "
         "If not set, dbt will look in the current working directory first, then HOME/.dbt/",
    nargs="?",
    default=os.getenv("DBT_PROFILES_DIR", default_profiles_dir()),
    type=str,
)

project_dir = CORE_PARSER.add_argument(
    "--project-dir",
    help="Which directory to look in for the dbt_project.yml file. "
         "Default is the current working directory and its parents.",
    nargs="?",
    default=os.getenv("DBT_PROJECT_DIR", default_project_dir()),
    type=str,
)

profile = CORE_PARSER.add_argument(
    "--profile",
    help="Which existing profile to load. Overrides setting in dbt_project.yml.",
    nargs="?",
    default=os.getenv("DBT_PROFILE"),
    type=str,
)

target = CORE_PARSER.add_argument(
    "--target",
    "-t",
    help="Which target to load for the given profile",
    nargs="?",
    default=os.getenv("DBT_TARGET"),
    type=str,
)

threads = CORE_PARSER.add_argument(
    "--threads",
    help="Specify number of threads to use while executing models. Overrides settings in profiles.yml.",
    nargs="?",
    default=None,
    type=int,
)

################################################################################
## DBT commands
################################################################################
CORE_PARSER.add_argument(
    "--clean",
    help="When this option is passed, run `dbt clean` before operations. "
         "If not passed, will attempt to load artifacts from the target folder before operations.",
    action='store_true'
)

CORE_PARSER.add_argument(
    "--deps",
    help="When this option is passed, run `dbt deps` before operations.",
    action='store_true'
)

################################################################################
## General runner args
################################################################################
CORE_PARSER.add_argument(
    "--config",
    help="Either the path to a contracts configuration file, "
         f"or the directory to look in for the {ContractsRunner.default_config_file_name!r} file. "
         "Defaults to the project dir when not specified.",
    nargs="?",
    default=None,
    type=Path,
)

CORE_PARSER.add_argument(
    "--contract",
    help="Run only this contract. If none given, apply all configured contracts. "
         "Specify granular contracts by separating keys by a '.' e.g. 'model', 'model.columns'",
    nargs="?",
    default=None,
    choices=[
        key for keys in (
            (contract.__config_key__, contract.child_config_key) for contract in CONTRACT_CLASSES
        ) for key in keys
    ],
    type=str,
)

CORE_PARSER.add_argument(
    "files",
    help="Apply contract to only these files. "
         "Must either be relative to the current folder, relative to the project folder, or absolute.",
    nargs="*",
    default=None,
    type=str,
)

################################################################################
## Validator runner args
################################################################################
VALIDATOR_PARSER = deepcopy(CORE_PARSER)
VALIDATOR_PARSER.prog = "dbt-validate"

VALIDATOR_PARSER.add_argument(
    "--output",
    help="Either the path to a file to write to when formatting results output, "
         f"or the directory to write a file to with filename {ContractsRunner.default_output_file_name!r}. "
         "Defaults to the project's target folder when not specified.",
    nargs="?",
    default=None,
    type=Path,
)

VALIDATOR_PARSER.add_argument(
    "--format",
    help="Specify the format of results output if desired. Output file will not be generated when not specified.",
    nargs="?",
    default=None,
    choices=ContractsRunner.output_writers_map.keys(),
    type=str,
)

VALIDATOR_PARSER.add_argument(
    "--no-fail",
    help="When this option is passed, do not fail when contracts do not pass.",
    action='store_true'
)

VALIDATOR_PARSER.add_argument(
    "--terms", "--term", "--validations", "--validation",
    help="Apply only these validations/terms. If none given, apply all configured validations/terms.",
    nargs="+",
    default=None,
    type=str,
)

################################################################################
## Generator runner args
################################################################################
GENERATOR_PARSER = deepcopy(CORE_PARSER)
GENERATOR_PARSER.prog = "dbt-generate"


def _setup_logging() -> None:
    logging.basicConfig(level=logging.INFO, format="%(message)s", handlers=[logging.StreamHandler()], force=True)


def _setup_runner(args: argparse.Namespace) -> ContractsRunner:
    args.profiles_dir = str(Path(args.profiles_dir).resolve())
    args.project_dir = str(Path(args.project_dir).resolve())

    config = get_config(args)

    if args.clean:
        clean_paths(config=config)
    if args.deps:
        install_dependencies(config=config)

    if args.config is None and args.project_dir:
        args.config = Path(args.project_dir)

    args.config = Path(args.config).resolve()

    runner = ContractsRunner.from_config(config)
    if args.files:
        runner.paths = config.args.files

    return runner


def clean() -> None:
    """Main entry point for executing `dbt clean`"""
    CORE_PARSER.prog = "dbt-clean"
    args, _ = CORE_PARSER.parse_known_args()
    config = get_config(args)
    clean_paths(config=config)


def deps() -> None:
    """Main entry point for executing `dbt deps`"""
    CORE_PARSER.prog = "dbt-deps"
    args, _ = CORE_PARSER.parse_known_args()
    config = get_config(args)
    install_dependencies(config=config)


def parse() -> None:
    """Main entry point for executing `dbt parse`"""
    CORE_PARSER.prog = "dbt-parse"
    args, _ = CORE_PARSER.parse_known_args()
    config = get_config(args)
    get_manifest(config=config)


def docs() -> None:
    """Main entry point for executing `dbt docs generate`"""
    CORE_PARSER.prog = "dbt-docs"
    args, _ = CORE_PARSER.parse_known_args()
    config = get_config(args)
    get_catalog(config=config)


def validate() -> None:
    """Main entry point for the `validator` CLI command"""
    _setup_logging()

    args, _ = VALIDATOR_PARSER.parse_known_args()
    runner = _setup_runner(args)

    if args.output is None:
        args.output = Path(runner.config.project_root, runner.config.target_path)
    args.output = Path(args.output).resolve()

    results = runner.validate(contract_key=args.contract, terms=args.terms)

    if args.format:
        runner.write_results(results, path=args.output, output_type=args.format)

    if not args.no_fail and results:
        raise Exception(f"Found {len(results)} contract violations.")


def generate() -> None:
    """Main entry point for the `generate` CLI command"""
    _setup_logging()

    args, _ = GENERATOR_PARSER.parse_known_args()
    runner = _setup_runner(args)

    runner.generate(contract_key=args.contract)


if __name__ == "__main__":
    operation = CORE_PARSER.add_argument(
        "--operation",
        help="Run this operation.",
        nargs="?",
        default="validate",
        choices=["validate", "generate"],
        type=str,
    )
    core_args, _ = CORE_PARSER.parse_known_args()

    if core_args.operation == "validate":
        validate()
        exit(0)
    elif core_args.operation == "generate":
        generate()
        exit(0)

    raise RuntimeError(f"Unrecognised operation: {core_args.command!r}")



import re
from abc import ABCMeta
from collections.abc import Iterable, Mapping, Collection, Sequence
from copy import copy
from typing import Any

from dbt.artifacts.resources.v1.components import ColumnInfo
from dbt.contracts.graph.manifest import Manifest
from dbt.contracts.graph.nodes import TestNode
from dbt_common.contracts.metadata import CatalogTable
from pydantic import Field, field_validator

from dbt_contracts.contracts._core import ContractContext
from dbt_contracts.contracts.matchers import StringMatcher, RangeMatcher
from dbt_contracts.contracts.terms._core import ChildContractTerm, validate_context
from dbt_contracts.contracts.utils import get_matching_catalog_table
from dbt_contracts.types import NodeT


class ColumnContractTerm[T: NodeT](ChildContractTerm[ColumnInfo, T], metaclass=ABCMeta):
    def _validate_node(self, column: ColumnInfo, node: NodeT, context: ContractContext) -> bool:
        has_column = column in node.columns.values()
        if not has_column:
            message = f"The column cannot be found in the {node.resource_type.lower()}"
            context.add_result(name=self.name, message=message, item=column, parent=node)

        return has_column

    def _get_and_validate_table(self, column: ColumnInfo, node: T, context: ContractContext) -> CatalogTable | None:
        table = get_matching_catalog_table(item=node, catalog=context.catalog)
        if table is None:
            message = f"The {node.resource_type.lower()} cannot be found in the database"
            context.add_result(name=self.name, message=message, item=column, parent=node)
            return

        has_column = column.name in table.columns.keys()
        if not has_column:
            message = f"The column cannot be found in the {table.metadata.type} {table.unique_id!r}"
            context.add_result(name=self.name, message=message, item=column, parent=node)
            return

        return table


class Exists[T: NodeT](ColumnContractTerm[T]):
    """Check whether the columns exist in the database."""
    needs_catalog = True

    @validate_context
    def run(self, item: ColumnInfo, context: ContractContext, parent: T = None) -> bool:
        if not self._validate_node(column=item, node=parent, context=context):
            return False

        return self._get_and_validate_table(column=item, node=parent, context=context) is not None


class HasTests[T: NodeT](ColumnContractTerm[T], RangeMatcher):
    """Check whether columns have an appropriate number of tests configured."""
    needs_manifest = True

    @staticmethod
    def _get_tests(column: ColumnInfo, node: NodeT, manifest: Manifest) -> Iterable[TestNode]:
        def _filter_nodes(test: Any) -> bool:
            return isinstance(test, TestNode) and all((
                test.attached_node == node.unique_id,
                test.column_name == column.name,
            ))

        return filter(_filter_nodes, manifest.nodes.values())

    @validate_context
    def run(self, item: ColumnInfo, context: ContractContext, parent: T = None) -> bool:
        if not self._validate_node(column=item, node=parent, context=context):
            return False

        count = len(tuple(self._get_tests(column=item, node=parent, manifest=context.manifest)))
        log_message = self._match(count=count, kind="tests")

        if log_message:
            context.add_result(name=self.name, message=log_message, item=item, parent=parent)
        return not log_message


class HasExpectedName[T: NodeT](ColumnContractTerm[T], StringMatcher):
    """
    Check whether columns have an expected name based on their data type.
    Matches the column names against a set of patterns mapped to specific data types.
    """
    patterns: Mapping[str, Sequence[str, ...]] = Field(
        description=(
            "A map of data types to regex patterns for which to "
            "validate names of columns which have the matching data type."
            "To define a generic contract which can apply to all unmatched data types, "
            "specify the data type key as `null`."
        ),
        default_factory=dict,
        examples=[{"BOOLEAN": ["(is|has|do)_.*"], "TIMESTAMP": [".*_at"], None: ["name_.*"]}]
    )

    # noinspection PyNestedDecorators
    @field_validator("patterns", mode="before")
    @classmethod
    def make_pattern_values_tuple(cls, patterns: Mapping[str, str | Sequence[str]]) -> dict[str, tuple[str]]:
        """Convert all meta values to tuples"""
        patterns = dict(copy(patterns))

        for key, val in patterns.items():
            if not isinstance(val, Collection) or isinstance(val, str):
                patterns[key] = (val,)
            else:
                patterns[key] = tuple(val)
        # noinspection PyTypeChecker
        return patterns

    def _get_column_data_type(self, column: ColumnInfo, node: T, context: ContractContext) -> str | None:
        if column.data_type:
            return column.data_type

        if context.catalog is None:
            return
        if (table := self._get_and_validate_table(column=column, node=node, context=context)) is None:
            return

        return table.columns[column.name].type

    @validate_context
    def run(self, item: ColumnInfo, context: ContractContext, parent: T = None) -> bool:
        if not self._validate_node(column=item, node=parent, context=context):
            return False
        if not (data_type := self._get_column_data_type(column=item, node=parent, context=context)):
            return False

        data_type = next((key for key in self.patterns if self._match(key, data_type)), None)
        patterns = self.patterns.get(data_type)
        if not patterns:  # no patterns defined for this data type
            return True

        expected_name = any(re.match(pattern, item.name) for pattern in patterns)
        if not expected_name:
            patterns_log = ', '.join(patterns)
            message = "Column name does not match expected patterns "
            message += f"for type {data_type!r}: {patterns_log}" if data_type else f": {patterns_log}"
            context.add_result(name=self.name, message=message, item=item, parent=parent)

        return expected_name


class HasDataType[T: NodeT](ColumnContractTerm[T]):
    """Check whether columns have a data type configured in their properties."""
    @validate_context
    def run(self, item: ColumnInfo, context: ContractContext, parent: T = None) -> bool:
        if not self._validate_node(column=item, node=parent, context=context):
            return False

        has_data_type = bool(item.data_type)
        if not has_data_type:
            message = "Data type not configured for this column"
            context.add_result(name=self.name, message=message, item=item, parent=parent)

        return has_data_type


class HasMatchingDescription[T: NodeT](ColumnContractTerm[T], StringMatcher):
    """Check whether the descriptions configured in columns' properties matches the descriptions in the database."""
    needs_catalog = True

    @validate_context
    def run(self, item: ColumnInfo, context: ContractContext, parent: T = None) -> bool:
        if not self._validate_node(column=item, node=parent, context=context):
            return False
        if (table := self._get_and_validate_table(column=item, node=parent, context=context)) is None:
            return False

        node_description = item.description
        table_description = table.columns[item.name].comment

        matched_description = self._match(node_description, table_description)
        if not matched_description:
            message = f"Description does not match remote entity: {node_description!r} != {table_description!r}"
            context.add_result(name=self.name, message=message, item=item, parent=parent)

        return matched_description


class HasMatchingDataType[T: NodeT](ColumnContractTerm[T], StringMatcher):
    """Check whether the data type configured in a column's properties matches the data type in the database."""
    needs_catalog = True

    @validate_context
    def run(self, item: ColumnInfo, context: ContractContext, parent: T = None) -> bool:
        if not self._validate_node(column=item, node=parent, context=context):
            return False
        if (table := self._get_and_validate_table(column=item, node=parent, context=context)) is None:
            return False

        node_type = item.data_type
        table_type = table.columns[item.name].type

        matched_type = self._match(node_type, table_type)
        if not matched_type:
            message = f"Data type does not match remote entity: {node_type!r} != {table_type!r}"
            context.add_result(name=self.name, message=message, item=item, parent=parent)

        return matched_type


class HasMatchingIndex[T: NodeT](ColumnContractTerm[T], StringMatcher):
    """
    Check whether the index position within the properties of a column's table
    matches the index position in the database.
    """
    needs_catalog = True

    @validate_context
    def run(self, item: ColumnInfo, context: ContractContext, parent: T = None) -> bool:
        if not self._validate_node(column=item, node=parent, context=context):
            return False
        if (table := self._get_and_validate_table(column=item, node=parent, context=context)) is None:
            return False

        node_index = list(parent.columns).index(item.name)
        table_index = table.columns[item.name].index

        matched_index = node_index == table_index
        if not matched_index:
            message = f"Column index does not match remote entity: {node_index} != {table_index}"
            context.add_result(name=self.name, message=message, item=item, parent=parent)

        return matched_index



from copy import deepcopy
from random import sample
from unittest import mock

import pytest
from _pytest.fixtures import FixtureRequest
from dbt.artifacts.resources import BaseResource
from dbt.contracts.graph.manifest import Manifest
from dbt.contracts.graph.nodes import TestNode, CompiledNode
from dbt_common.contracts.metadata import CatalogTable
from faker import Faker

from dbt_contracts.contracts import ContractContext
from dbt_contracts.contracts.terms.node import Exists, HasTests, HasAllColumns, HasExpectedColumns, \
    HasMatchingDescription, HasContract, HasValidUpstreamDependencies, HasValidRefDependencies, \
    HasValidSourceDependencies, HasValidMacroDependencies, HasNoFinalSemicolon, HasNoHardcodedRefs
from dbt_contracts.contracts.utils import get_matching_catalog_table


def test_exists(node: CompiledNode, simple_resource: BaseResource, context: ContractContext):
    assert get_matching_catalog_table(item=node, catalog=context.catalog) is not None
    assert Exists().run(node, context=context)

    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        # noinspection PyTypeChecker
        assert not Exists().run(simple_resource, context=context)
        mock_add_result.assert_called_once()


def test_get_tests(node: CompiledNode, simple_resource: BaseResource, manifest: Manifest):
    tests = list(HasTests._get_tests(item=node, manifest=manifest))
    assert tests
    assert all(isinstance(test, TestNode) for test in tests)
    assert all(test.attached_node == node.unique_id for test in tests)

    # noinspection PyTypeChecker
    assert not list(HasTests._get_tests(item=simple_resource, manifest=manifest))


def test_has_tests(node: CompiledNode, simple_resource: BaseResource, context: ContractContext):
    assert 0 < len(list(HasTests._get_tests(item=node, manifest=context.manifest))) < 10
    assert HasTests().run(node, context=context)

    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasTests(min_count=10).run(node, context=context)
        mock_add_result.assert_called_once()

        # noinspection PyTypeChecker
        assert not HasTests().run(simple_resource, context=context)
        assert len(mock_add_result.mock_calls) == 2


# noinspection PyTestUnpassedFixture
def test_has_all_columns(node: CompiledNode, catalog_table: CatalogTable, context: ContractContext):
    assert set(node.columns.keys()) == set(catalog_table.columns.keys())
    assert HasAllColumns().run(node, context=context)  # fixtures are set up to match

    node.columns.clear()
    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasAllColumns().run(node, context=context)
        mock_add_result.assert_called_once()


# noinspection PyTestUnpassedFixture
def test_has_expected_column_names(node: CompiledNode, context: ContractContext, faker: Faker):
    assert HasExpectedColumns().run(node, context=context)  # no expectations set
    assert HasExpectedColumns(columns=sample(list(node.columns), k=2)).run(node, context=context)

    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasExpectedColumns(columns=faker.words(10)).run(node, context=context)
        mock_add_result.assert_called_once()


# noinspection PyTestUnpassedFixture
def test_has_expected_column_types(node: CompiledNode, context: ContractContext):
    columns = {column.name: column.data_type for column in node.columns.values() if column.data_type}
    assert HasExpectedColumns(columns=columns).run(node, context=context)

    columns = {column.name: "incorrect type" for column in node.columns.values()}
    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasExpectedColumns(columns=columns).run(node, context=context)
        mock_add_result.assert_called_once()


def test_has_matching_description(
        node: CompiledNode, catalog_table: CatalogTable, context: ContractContext, faker: Faker
):
    node.description = catalog_table.metadata.comment
    assert HasMatchingDescription().run(node, context=context)  # fixtures are set up to match

    node.description = faker.sentence()
    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasMatchingDescription().run(node, context=context)
        mock_add_result.assert_called_once()


# noinspection PyTestUnpassedFixture
@pytest.fixture(params=["model"])
def compiled_node(request: FixtureRequest) -> CompiledNode:
    """CompiledNode fixture"""
    return deepcopy(request.getfixturevalue(request.param))


# noinspection PyTestUnpassedFixture
@pytest.fixture
def contract_node(compiled_node: CompiledNode) -> CompiledNode:
    """CompiledNode fixture with enforced contract"""
    compiled_node.contract.enforced = True
    for column in compiled_node.columns.values():
        column.data_type = "int"

    return compiled_node


def test_has_valid_contract(contract_node: CompiledNode, context: ContractContext):
    with mock.patch.object(HasAllColumns, "run", return_value=True) as mock_has_all_columns:
        assert HasContract().run(contract_node, context=context)
        mock_has_all_columns.assert_called_once_with(contract_node, context=context)


def test_has_invalid_contract_not_enforced(contract_node: CompiledNode, context: ContractContext):
    contract_node.contract.enforced = False

    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasContract().run(contract_node, context=context)
        mock_add_result.assert_called_once()
        assert mock_add_result.mock_calls[0].kwargs["message"] == "Contract not enforced"


def test_has_invalid_contract_missing_data_types(contract_node: CompiledNode, context: ContractContext):
    for column in sample(list(contract_node.columns.values()), k=2):
        column.data_type = None

    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasContract().run(contract_node, context=context)
        mock_add_result.assert_called_once()
        assert "all data types must be declared" in mock_add_result.mock_calls[0].kwargs["message"]


def test_add_result_for_invalid_dependencies(compiled_node: CompiledNode, context: ContractContext):
    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasValidUpstreamDependencies._add_result_for_invalid_dependencies(
            item=compiled_node, kind="ref", context=context, missing=["1", "2", "3"]
        )
        mock_add_result.assert_called_once()


def test_has_valid_ref_dependencies(compiled_node: CompiledNode, context: ContractContext):
    assert not compiled_node.depends_on_nodes
    assert HasValidRefDependencies().run(compiled_node, context=context)

    upstream_deps = sample([node for node in context.manifest.nodes if node.startswith("model")], k=3)
    assert upstream_deps
    compiled_node.depends_on_nodes.extend(upstream_deps)
    assert HasValidRefDependencies().run(compiled_node, context=context)

    compiled_node.depends_on_nodes.append("model.invalid_ref")
    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasValidRefDependencies().run(compiled_node, context=context)
        mock_add_result.assert_called_once()


def test_has_valid_source_dependencies(compiled_node: CompiledNode, context: ContractContext):
    assert not compiled_node.depends_on_nodes
    assert HasValidSourceDependencies().run(compiled_node, context=context)

    upstream_deps = sample(list(context.manifest.sources), k=3)
    assert upstream_deps
    compiled_node.depends_on_nodes.extend(upstream_deps)
    assert HasValidSourceDependencies().run(compiled_node, context=context)

    compiled_node.depends_on_nodes.append("source.invalid_ref")
    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasValidSourceDependencies().run(compiled_node, context=context)
        mock_add_result.assert_called_once()


def test_has_valid_macro_dependencies(compiled_node: CompiledNode, context: ContractContext):
    assert not compiled_node.depends_on_macros
    assert HasValidMacroDependencies().run(compiled_node, context=context)

    upstream_deps = sample(list(context.manifest.macros), k=3)
    assert upstream_deps
    compiled_node.depends_on_macros.extend(upstream_deps)
    assert HasValidMacroDependencies().run(compiled_node, context=context)

    compiled_node.depends_on_macros.append("macro.invalid_ref")
    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasValidMacroDependencies().run(compiled_node, context=context)
        mock_add_result.assert_called_once()


def test_has_no_final_semicolon(compiled_node: CompiledNode, context: ContractContext, faker: Faker):
    compiled_node.path = faker.file_path(extension="sql", absolute=False)
    compiled_node.raw_code = "SELECT * FROM table"
    assert HasNoFinalSemicolon().run(compiled_node, context=context)

    compiled_node.raw_code = "SELECT * FROM table \n;   \n \n "
    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasNoFinalSemicolon().run(compiled_node, context=context)
        mock_add_result.assert_called_once()


def test_has_no_hardcoded_refs(compiled_node: CompiledNode, context: ContractContext, faker: Faker):
    compiled_node.path = faker.file_path(extension="sql", absolute=False)
    compiled_node.raw_code = """
        WITH cte1 as (SELECT * FROM VALUES(1,2,3)), cte2 as (SELECT * FROM VALUES(1,2,3))
        SELECT * FROM {{ ref('model_ref')}} ref1
        JOIN cte1 ON cte1.a = ref.a
        JOIN cte2 ON cte2.a = ref.a
    """
    assert HasNoHardcodedRefs().run(compiled_node, context=context)

    compiled_node.raw_code = "SELECT * FROM table"
    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasNoHardcodedRefs().run(compiled_node, context=context)
        mock_add_result.assert_called_once()



from dbt.contracts.graph.nodes import SourceDefinition

from dbt_contracts.contracts._core import ContractContext
from dbt_contracts.contracts.matchers import RangeMatcher
from dbt_contracts.contracts.terms._core import validate_context
from dbt_contracts.contracts.terms.node import NodeContractTerm


class HasLoader(NodeContractTerm[SourceDefinition]):
    """Check whether sources have appropriate configuration for a loader in their properties."""
    @validate_context
    def run(self, item: SourceDefinition, context: ContractContext) -> bool:
        has_loader = bool(item.loader)
        if not has_loader:
            message = "Loader is not correctly configured"
            context.add_result(name=self.name, message=message, item=item)

        return has_loader


class HasFreshness(NodeContractTerm[SourceDefinition]):
    """Check whether sources have freshness configured in their properties."""
    @validate_context
    def run(self, item: SourceDefinition, context: ContractContext) -> bool:
        has_freshness = bool(item.loaded_at_field) and item.has_freshness
        if not has_freshness:
            message = "Freshness is not correctly configured"
            context.add_result(name=self.name, message=message, item=item)

        return has_freshness


class HasDownstreamDependencies(NodeContractTerm[SourceDefinition], RangeMatcher):
    """Check whether sources have an appropriate number of downstream dependencies."""
    needs_manifest = True

    @validate_context
    def run(self, item: SourceDefinition, context: ContractContext) -> bool:
        count = sum(item.unique_id in node.depends_on_nodes for node in context.manifest.nodes.values())
        log_message = self._match(count=count, kind="downstream dependencies")

        if log_message:
            context.add_result(name=self.name, message=log_message, item=item)
        return not log_message



from collections.abc import Sequence, Collection, Mapping
from copy import copy
from typing import Annotated

from dbt.contracts.graph.nodes import SourceDefinition
from pydantic import BeforeValidator, Field, field_validator

from dbt_contracts.contracts._core import ContractContext
from dbt_contracts.contracts.terms._core import ParentContractTerm, ChildContractTerm, validate_context
from dbt_contracts.contracts.utils import to_tuple
from dbt_contracts.types import ParentT, PropertiesT, DescriptionT, TagT, MetaT


class HasProperties[I: PropertiesT](ParentContractTerm[I]):
    """Check whether the {kind} have properties files defined."""
    @validate_context
    def run(self, item: I, context: ContractContext) -> bool:
        if isinstance(item, SourceDefinition):  # sources always have properties files defined
            return True

        has_properties = bool(item.patch_path)
        if not has_properties:
            message = "No properties file found"
            context.add_result(name=self.name, message=message, item=item)

        return has_properties


class HasDescription[I: DescriptionT, P: ParentT](ChildContractTerm[I, P]):
    """Check whether the {kind} have descriptions defined in their properties."""
    @validate_context
    def run(self, item: I, context: ContractContext, parent: P = None) -> bool:
        has_description = bool(item.description)
        if not has_description:
            message = "Missing description"
            context.add_result(name=self.name, message=message, item=item, parent=parent)

        return has_description


class HasRequiredTags[I: TagT, P: ParentT](ChildContractTerm[I, P]):
    """Check whether the {kind} have the expected set of required tags set."""
    tags: Annotated[Sequence[str], BeforeValidator(to_tuple)] = Field(
        description="The required tags",
        default=tuple(),
        examples=["tag1", ["tag1", "tag2"]],
    )

    @validate_context
    def run(self, item: I, context: ContractContext, parent: P = None) -> bool:
        missing_tags = set(self.tags) - set(item.tags)
        if missing_tags:
            message = f"Missing required tags: {', '.join(missing_tags)}"
            context.add_result(name=self.name, message=message, item=item, parent=parent)

        return not missing_tags


class HasAllowedTags[I: TagT, P: ParentT](ChildContractTerm[I, P]):
    """Check whether the {kind} have only tags set from a configured permitted list."""
    tags: Annotated[Sequence[str], BeforeValidator(to_tuple)] = Field(
        description="The allowed tags",
        default=tuple(),
        examples=["tag1", ["tag1", "tag2"]],
    )

    @validate_context
    def run(self, item: I, context: ContractContext, parent: P = None) -> bool:
        invalid_tags = set(item.tags) - set(self.tags)
        if invalid_tags:
            message = f"Contains invalid tags: {', '.join(invalid_tags)}"
            context.add_result(name=self.name, message=message, item=item, parent=parent)

        return len(invalid_tags) == 0


class HasRequiredMetaKeys[I: MetaT, P: ParentT](ChildContractTerm[I, P]):
    """Check whether the {kind} have the expected set of required meta keys set."""
    keys: Annotated[Sequence[str], BeforeValidator(to_tuple)] = Field(
        description="The required meta keys",
        default=tuple(),
        examples=["key1", ["key1", "key2"]],
    )

    @validate_context
    def run(self, item: I, context: ContractContext, parent: P = None) -> bool:
        missing_keys = set(self.keys) - set(item.meta.keys())
        if missing_keys:
            message = f"Missing required keys: {', '.join(missing_keys)}"
            context.add_result(name=self.name, message=message, item=item, parent=parent)

        return not missing_keys


class HasAllowedMetaKeys[I: MetaT, P: ParentT](ChildContractTerm[I, P]):
    """Check whether the {kind} have only meta keys set from a configured permitted list."""
    keys: Annotated[Sequence[str], BeforeValidator(to_tuple)] = Field(
        description="The allowed meta keys",
        default=tuple(),
        examples=["key1", ["key1", "key2"]],
    )

    @validate_context
    def run(self, item: I, context: ContractContext, parent: P = None) -> bool:
        invalid_keys = set(item.meta.keys()) - set(self.keys)
        if invalid_keys:
            message = f"Contains invalid keys: {', '.join(invalid_keys)}"
            context.add_result(name=self.name, message=message, item=item, parent=parent)

        return len(invalid_keys) == 0


class HasAllowedMetaValues[I: MetaT, P: ParentT](ChildContractTerm[I, P]):
    """Check whether the {kind} have only meta values set from a configured permitted mapping of keys to values."""
    meta: Mapping[str, Sequence[str]] = Field(
        description="The mapping of meta keys to their allowed values",
        default_factory=dict,
        examples=[{"key1": "val1", "key2": ["val2", "val3"]}],
    )

    # noinspection PyNestedDecorators
    @field_validator("meta", mode="before")
    @classmethod
    def make_meta_values_tuple(cls, meta: Mapping[str, str | Sequence[str]]) -> dict[str, tuple[str]]:
        """Convert all meta values to tuples"""
        meta = dict(copy(meta))

        for key, val in meta.items():
            if not isinstance(val, Collection) or isinstance(val, str):
                meta[key] = (val,)
            else:
                meta[key] = tuple(val)
        # noinspection PyTypeChecker
        return meta

    @validate_context
    def run(self, item: I, context: ContractContext, parent: P = None) -> bool:
        invalid_meta: dict[str, str] = {}
        expected_meta: dict[str, Collection[str]] = {}

        for key, values in self.meta.items():
            if not isinstance(values, Collection) or isinstance(values, str):
                values = [values]
            if key in item.meta and item.meta[key] not in values:
                invalid_meta[key] = item.meta[key]
                expected_meta[key] = values

        if invalid_meta:
            message = f"Contains invalid meta values: {invalid_meta} | Accepted values: {expected_meta}"
            context.add_result(name=self.name, message=message, item=item, parent=parent)

        return not invalid_meta



import re
from abc import ABCMeta
from collections.abc import Iterable, Sequence, Mapping, Collection, Generator
from pathlib import Path
from typing import Any, ClassVar

from dbt.contracts.graph.manifest import Manifest
from dbt.contracts.graph.nodes import TestNode, CompiledNode
from pydantic import Field

from dbt_contracts.contracts._core import ContractContext
from dbt_contracts.contracts.matchers import RangeMatcher, StringMatcher
from dbt_contracts.contracts.terms._core import ParentContractTerm, validate_context
from dbt_contracts.contracts.utils import get_matching_catalog_table
from dbt_contracts.types import NodeT


class NodeContractTerm[T: NodeT](ParentContractTerm[T], metaclass=ABCMeta):
    pass


class Exists[T: NodeT](NodeContractTerm[T]):
    """Check whether the {kind} exist in the database."""
    needs_catalog = True

    @validate_context
    def run(self, item: T, context: ContractContext) -> bool:
        table = get_matching_catalog_table(item, catalog=context.catalog)
        if table is None:
            message = f"The {item.resource_type.lower()} cannot be found in the database"
            context.add_result(name=self.name, message=message, item=item)

        return table is not None


class HasTests[T: NodeT](NodeContractTerm[T], RangeMatcher):
    """Check whether {kind} have an appropriate number of tests configured."""
    needs_manifest = True

    @staticmethod
    def _get_tests(item: NodeT, manifest: Manifest) -> Iterable[TestNode]:
        def _filter_nodes(test: Any) -> bool:
            return isinstance(test, TestNode) and all((
                test.attached_node == item.unique_id,
                test.column_name is None,
            ))

        return filter(_filter_nodes, manifest.nodes.values())

    @validate_context
    def run(self, item: T, context: ContractContext) -> bool:
        count = len(tuple(self._get_tests(item, manifest=context.manifest)))
        log_message = self._match(count=count, kind="tests")

        if log_message:
            context.add_result(name=self.name, message=log_message, item=item)
        return not log_message


class HasAllColumns[T: NodeT](NodeContractTerm[T]):
    """
    Check whether {kind} have all columns set in their properties.
    Ensures that all columns present in the database are present in dbt project properties.
    """
    needs_catalog = True

    @validate_context
    def run(self, item: T, context: ContractContext) -> bool:
        table = get_matching_catalog_table(item, catalog=context.catalog)
        if not table:
            return False

        actual_columns = {column.name for column in item.columns.values()}
        expected_columns = {column.name for column in table.columns.values()}

        missing_columns = expected_columns - actual_columns
        if missing_columns:
            message = (
                f"{item.resource_type.title()} config does not contain all columns. "
                f"Missing {', '.join(missing_columns)}"
            )
            context.add_result(name=self.name, message=message, item=item)

        extra_columns = actual_columns - expected_columns
        if extra_columns:
            message = (
                f"{item.resource_type.title()} config contains too many columns. "
                f"Extra {', '.join(extra_columns)}"
            )
            context.add_result(name=self.name, message=message, item=item)

        return not missing_columns and not extra_columns


class HasExpectedColumns[T: NodeT](NodeContractTerm[T]):
    """
    Check whether {kind} have the expected names of columns set in their properties.
    Also checks if those columns have the expected data types if configured to do so.
    """
    columns: str | Sequence[str] | Mapping[str, str] = Field(
        description="A sequence of the names of the columns that should exist in the node, "
                    "or a mapping of the column names and their associated data types that should exist.",
        default=tuple(),
        examples=["column1", ["column1", "column2", "column3"], {"column1": "VARCHAR", "column2": "INT"}]
    )

    @validate_context
    def run(self, item: T, context: ContractContext) -> bool:
        node_column_types = {column.name: column.data_type for column in item.columns.values()}

        missing_columns = set()
        if self.columns:
            missing_columns = set(self.columns) - set(node_column_types)
        if missing_columns:
            message = (
                f"{item.resource_type.title()} does not have all expected columns. "
                f"Missing: {', '.join(missing_columns)}"
            )
            context.add_result(name=self.name, message=message, item=item)

        unexpected_types = {}
        if isinstance(self.columns, Mapping):
            unexpected_types = {
                name: (node_column_types[name], data_type) for name, data_type in self.columns.items()
                if name in node_column_types and node_column_types[name] != data_type
            }
        if unexpected_types:
            message = f"{item.resource_type.title()} has unexpected column types."
            for name, (actual, expected) in unexpected_types.items():
                message += f"\n- {actual!r} should be {expected!r}"

            context.add_result(name=self.name, message=message, item=item)

        return not missing_columns and not unexpected_types


class HasMatchingDescription[T: NodeT](NodeContractTerm[T], StringMatcher):
    """Check whether the descriptions configured in {kind}' properties match the descriptions in the database."""
    needs_catalog = True

    @validate_context
    def run(self, item: T, context: ContractContext) -> bool:
        table = get_matching_catalog_table(item, catalog=context.catalog)
        if not table:
            return False

        matched_description = self._match(item.description, table.metadata.comment)
        if not matched_description:
            message = f"Description does not match remote entity: {item.description!r} != {table.metadata.comment!r}"
            context.add_result(name=self.name, message=message, item=item)

        return matched_description


class HasContract[T: CompiledNode](NodeContractTerm[T]):
    """Check whether {kind} have appropriate configuration for a contract in their properties."""
    @validate_context
    def run(self, item: T, context: ContractContext) -> bool:
        has_contract = bool(item.contract.enforced)
        if not has_contract:
            message = "Contract not enforced"
            context.add_result(name=self.name, message=message, item=item)

        # node must have all columns defined for contract to be valid
        has_all_columns = HasAllColumns().run(item, context=context)

        has_data_types = all(column.data_type for column in item.columns.values())
        if not has_data_types:
            message = "To enforce a contract, all data types must be declared"
            context.add_result(name=self.name, message=message, item=item)

        return all((has_contract, has_all_columns, has_data_types))


class HasValidUpstreamDependencies[T: CompiledNode](NodeContractTerm[T], metaclass=ABCMeta):
    needs_manifest = True

    @staticmethod
    def _add_result_for_invalid_dependencies(
            item: T, kind: str, context: ContractContext, missing: Collection
    ) -> bool:
        if missing:
            kind = kind.rstrip("s")
            message = (
                f"{item.resource_type.title()} has missing upstream {kind} dependencies declared: "
                f"{', '.join(missing)}"
            )
            context.add_result(name=f"has_valid_{kind}_dependencies", message=message, item=item)

        return not missing


class HasValidRefDependencies[T: CompiledNode](HasValidUpstreamDependencies[T]):
    """
    Check whether {kind} have an appropriate number of upstream dependencies
    i.e. the number of `ref` macros present in the query.
    """
    @validate_context
    def run(self, item: T, context: ContractContext) -> bool:
        upstream_dependencies = {ref for ref in item.depends_on_nodes if ref.startswith("model")}
        missing_dependencies = upstream_dependencies - set(context.manifest.nodes.keys())

        return self._add_result_for_invalid_dependencies(
            item, kind="ref", context=context, missing=missing_dependencies
        )


class HasValidSourceDependencies[T: CompiledNode](HasValidUpstreamDependencies[T]):
    """
    Check whether {kind} have an appropriate number of upstream dependencies for sources
    i.e. the number of `source` macros present in the query.
    """
    @validate_context
    def run(self, item: T, context: ContractContext) -> bool:
        upstream_dependencies = {ref for ref in item.depends_on_nodes if ref.startswith("source")}
        missing_dependencies = upstream_dependencies - set(context.manifest.sources.keys())

        return self._add_result_for_invalid_dependencies(
            item, kind="source", context=context, missing=missing_dependencies
        )


class HasValidMacroDependencies[T: CompiledNode](HasValidUpstreamDependencies[T]):
    """
    Check whether {kind} have an appropriate number of upstream dependencies for macros
    i.e. the number of custom macros present in the query.
    """
    @validate_context
    def run(self, item: T, context: ContractContext) -> bool:
        upstream_dependencies = set(item.depends_on_macros)
        missing_dependencies = upstream_dependencies - set(context.manifest.macros.keys())

        return self._add_result_for_invalid_dependencies(
            item, kind="macro", context=context, missing=missing_dependencies
        )


class HasNoFinalSemicolon[T: CompiledNode](NodeContractTerm[T]):
    """Check if {kind} have a final semicolon present in their queries."""
    @validate_context
    def run(self, item: T, context: ContractContext) -> bool:
        # ignore non-SQL models
        if Path(item.path).suffix.casefold() != ".sql":
            return True

        has_final_semicolon = item.raw_code.strip().endswith(";")
        if has_final_semicolon:
            message = "Script has a final semicolon"
            context.add_result(name=self.name, message=message, item=item)

        return not has_final_semicolon


class HasNoHardcodedRefs[T: CompiledNode](NodeContractTerm[T]):
    """Check if {kind} have any hardcoded references to database objects in their queries."""
    comments_pattern: ClassVar[str] = r"(?<=(\/\*|\{#))((.|[\r\n])+?)(?=(\*+\/|#\}))|[ \t]*--.*"

    cte_keywords: ClassVar[frozenset[str]] = frozenset({"with"})
    cte_pattern: ClassVar[str] = r"^[\w\d_-]+$"

    ref_keywords: ClassVar[frozenset[str]] = frozenset({"from", "join"})
    ref_ignore: ClassVar[frozenset[str]] = frozenset({"values"})

    @classmethod
    def _remove_comments(cls, script: str) -> str:
        return re.sub(cls.comments_pattern, "", script)

    @classmethod
    def _add_spacing(cls, script: str) -> str:
        script = re.sub(r"\{\s*\{", "{{ ", script)
        script = re.sub(r"}\s*}", " }}", script)
        script = re.sub(r"([()])", r" \1 ", script)
        return script

    @classmethod
    def _iter_script_tokens(cls, script: str) -> Generator[tuple[str, str, str]]:
        sql = script.split(";")[0]
        sql = cls._remove_comments(sql)
        sql = cls._add_spacing(sql)
        tokens = iter(sql.split())

        prev_token = None
        curr_token = next(tokens, None)
        if curr_token is not None:
            curr_token = curr_token.casefold()

        while (next_token := next(tokens, None)) is not None:
            next_token = next_token.casefold()
            yield prev_token, curr_token, next_token
            prev_token = curr_token
            curr_token = next_token

        yield prev_token, curr_token, None

    def _get_ref(self, prev_token: str, curr_token: str) -> str | None:
        if prev_token not in self.ref_keywords:
            return
        if curr_token in self.ref_ignore:
            return
        if any(curr_token.startswith(char) for char in {"{", "("}):
            return
        return curr_token.strip(",")

    def _get_cte(self, prev_token: str, curr_token: str, next_token: str) -> str | None:
        if prev_token in self.cte_keywords and re.match(self.cte_pattern, curr_token, re.I):
            return curr_token

        if curr_token == "as" and next_token.startswith("(") and re.match(self.cte_pattern, prev_token, re.I):
            return prev_token

    @validate_context
    def run(self, item: T, context: ContractContext) -> bool:
        # ignore non-SQL models
        if Path(item.path).suffix.casefold() != ".sql":
            return True

        refs = set()
        ctes = set()
        for prev_token, curr_token, next_token in self._iter_script_tokens(item.raw_code):
            if not prev_token or not next_token:
                continue

            if ref := self._get_ref(curr_token, next_token):
                refs.add(ref)
            elif cte := self._get_cte(prev_token, curr_token, next_token):
                ctes.add(cte)

        hardcoded_refs = refs - ctes
        if hardcoded_refs:
            message = f"Script has hardcoded refs: {', '.join(hardcoded_refs)}"
            context.add_result(name=self.name, message=message, item=item)

        return not hardcoded_refs



from copy import copy, deepcopy
from random import choice
from unittest import mock

import pytest
from dbt.artifacts.resources import BaseResource
from dbt.artifacts.resources.v1.components import ColumnInfo
from dbt.contracts.graph.manifest import Manifest
from dbt.contracts.graph.nodes import TestNode, ModelNode, CompiledNode
from dbt_common.contracts.metadata import CatalogTable
from faker import Faker

from dbt_contracts.contracts import ContractContext
from dbt_contracts.contracts.terms.column import ColumnContractTerm, Exists, HasTests, HasExpectedName, HasDataType, \
    HasMatchingDescription, HasMatchingDataType, HasMatchingIndex
from dbt_contracts.contracts.utils import get_matching_catalog_table
from dbt_contracts.types import NodeT


# noinspection PyTestUnpassedFixture
def test_validate_node(node: NodeT, node_column: ColumnInfo, context: ContractContext):
    assert Exists()._validate_node(node=node, column=node_column, context=context)

    node.columns.pop(node_column.name)
    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not Exists()._validate_node(node=node, column=node_column, context=context)
        mock_add_result.assert_called_once()


def test_get_and_validate_table(
        node: NodeT, node_column: ColumnInfo, catalog_table: CatalogTable, context: ContractContext
):
    assert Exists()._get_and_validate_table(node=node, column=node_column, context=context) == catalog_table


def test_get_and_validate_table_unmatched_table(
        node: NodeT, node_column: ColumnInfo, catalog_table: CatalogTable, context: ContractContext
):
    node.unique_id = "unknown_id"
    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert Exists()._get_and_validate_table(node=node, column=node_column, context=context) is None
        mock_add_result.assert_called_once()


# noinspection PyTestUnpassedFixture
def test_get_and_validate_table_missing_column(
        node: NodeT, node_column: ColumnInfo, catalog_table: CatalogTable, context: ContractContext
):
    # add a new column to the node to simulate a column being remove
    # preferred way to set up test as the catalog_table should not be modified to ensure other tests pass
    new_column = deepcopy(node_column)
    new_column.name = "new_columns"
    node.columns[new_column.name] = new_column

    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert Exists()._get_and_validate_table(node=node, column=new_column, context=context) is None
        mock_add_result.assert_called_once()


@pytest.mark.parametrize("term", [
    Exists(),
    HasTests(),
    HasExpectedName(),
    HasDataType(),
    HasMatchingDescription(),
    HasMatchingDataType(),
    HasMatchingIndex()
])
def test_terms_validate_column_in_node(term: ColumnContractTerm, model: ModelNode, context: ContractContext):
    column = choice(list(model.columns.values()))

    with mock.patch.object(term.__class__, "_validate_node", return_value=False) as mock_validate_node:
        assert not term.run(item=column, parent=model, context=context)
        mock_validate_node.assert_called_once()

    model.columns.pop(column.name)
    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not term.run(item=column, parent=model, context=context)
        mock_add_result.assert_called_once()
        assert "The column cannot be found" in mock_add_result.mock_calls[0].kwargs["message"]


def test_exists(node: CompiledNode, node_column: ColumnInfo, context: ContractContext):
    assert get_matching_catalog_table(item=node, catalog=context.catalog) is not None
    assert Exists().run(node_column, parent=node, context=context)

    node.columns.pop(node_column.name)
    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not Exists().run(node_column, parent=node, context=context)
        mock_add_result.assert_called_once()


# noinspection PyTestUnpassedFixture
def test_get_tests(node: NodeT, node_column: ColumnInfo, simple_resource: BaseResource, manifest: Manifest):
    tests = list(HasTests._get_tests(column=node_column, node=node, manifest=manifest))
    assert tests
    assert all(isinstance(test, TestNode) for test in tests)
    assert all(test.attached_node == node.unique_id and test.column_name == node_column.name for test in tests)

    # noinspection PyTypeChecker
    assert not list(HasTests._get_tests(column=node_column, node=simple_resource, manifest=manifest))


def test_has_tests(node: CompiledNode, node_column: ColumnInfo, context: ContractContext):
    assert 0 < len(list(HasTests._get_tests(column=node_column, node=node, manifest=context.manifest))) < 10
    assert HasTests().run(node_column, parent=node, context=context)

    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasTests(min_count=10).run(node_column, parent=node, context=context)
        mock_add_result.assert_called_once()


def test_get_column_data_type(
        node: CompiledNode, node_column: ColumnInfo, catalog_table: CatalogTable, context: ContractContext
):
    node_column.data_type = "int"
    catalog_table.columns[node_column.name].type = "str"

    # gets data type from node
    assert HasExpectedName()._get_column_data_type(column=node_column, node=node, context=context) == "int"

    # gets data type from table when node data type is not available
    node_column.data_type = None
    assert HasExpectedName()._get_column_data_type(column=node_column, node=node, context=context) == "str"

    # returns safely when table or catalog are not available
    node.unique_id = "unknown_id"
    assert HasExpectedName()._get_column_data_type(column=node_column, node=node, context=context) is None
    context = copy(context)
    assert HasExpectedName()._get_column_data_type(column=node_column, node=node, context=context) is None


def test_has_expected_name(
        node: CompiledNode, node_column: ColumnInfo, catalog_table: CatalogTable, context: ContractContext
):
    node_column.data_type = "boolean"
    # always returns true when data type is defined in patterns
    assert HasExpectedName(patterns={"str": ".*"}).run(node_column, parent=node, context=context)

    node_column.name = "has_name"
    assert HasExpectedName(patterns={"boolean": "has_.*"}).run(node_column, parent=node, context=context)

    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasExpectedName(patterns={"boolean": "is_.*"}).run(node_column, parent=node, context=context)
        mock_add_result.assert_called_once()


def test_has_data_type(
        node: CompiledNode, node_column: ColumnInfo, catalog_table: CatalogTable, context: ContractContext
):
    node_column.data_type = "int"
    assert HasDataType().run(node_column, parent=node, context=context)

    node_column.data_type = None
    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasDataType().run(node_column, parent=node, context=context)
        mock_add_result.assert_called_once()


def test_has_matching_description(
        node: CompiledNode, node_column: ColumnInfo, catalog_table: CatalogTable, faker: Faker, context: ContractContext
):
    node_column.description = catalog_table.columns[node_column.name].comment
    assert HasMatchingDescription().run(node_column, parent=node, context=context)

    node_column.description = faker.sentence()
    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasMatchingDescription().run(node_column, parent=node, context=context)
        mock_add_result.assert_called_once()


def test_has_matching_data_type(
        node: CompiledNode, node_column: ColumnInfo, catalog_table: CatalogTable, faker: Faker, context: ContractContext
):
    node_column.data_type = catalog_table.columns[node_column.name].type
    assert HasMatchingDataType().run(node_column, parent=node, context=context)

    node_column.data_type = "_".join(faker.words())
    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasMatchingDataType().run(node_column, parent=node, context=context)
        mock_add_result.assert_called_once()


def test_has_matching_index(
        node: CompiledNode, node_column: ColumnInfo, catalog_table: CatalogTable, context: ContractContext
):
    assert list(node.columns).index(node_column.name) == catalog_table.columns[node_column.name].index
    assert HasMatchingIndex().run(node_column, parent=node, context=context)

    # reorder columns
    if list(node.columns.keys()).index(node_column.name) == (len(node.columns) - 1):
        first_column = node.columns.pop(list(node.columns.keys())[0])
        node.columns[first_column.name] = first_column
    else:
        node.columns.pop(node_column.name)
        node.columns[node_column.name] = node_column

    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasMatchingIndex().run(node_column, parent=node, context=context)
        mock_add_result.assert_called_once()



from __future__ import annotations

import re
from abc import ABCMeta, abstractmethod
from collections.abc import Callable
from typing import ClassVar, Union

from dbt.artifacts.schemas.catalog import CatalogArtifact
from dbt.contracts.graph.manifest import Manifest

from dbt_contracts.contracts._core import ContractContext, ContractPart
from dbt_contracts.types import ItemT, ChildT, ParentT


class ContractTerm[I: ItemT](ContractPart, metaclass=ABCMeta):
    """
    A part of a contract meant to apply checks on a specific item according to a set of rules.

    May also process an item while also taking into account its parent item
    e.g. a Column (child item) within a Model (parent item)
    """
    #: Mark if this term requires a loaded manifest to operate.
    #: The manifest must then be provided by the ContractContext for the term to execute successfully.
    needs_manifest: ClassVar[bool] = False

    #: Mark if this term requires a loaded catalog to operate.
    #: The catalog must then be provided by the ContractContext for the term to execute successfully.
    needs_catalog: ClassVar[bool] = False

    @classmethod
    def _name(cls) -> str:
        """The name of this term in snake_case."""
        class_name = cls.__name__.replace("Term", "")
        return re.sub(r"([a-z])([A-Z])", r"\1_\2", class_name).lower()

    @abstractmethod
    def run(self, item: I, context: ContractContext) -> bool:
        """
        Run this term on the given item.

        :param item: The item to check.
        :param context: The contract context to use.
        :return: Boolean for if the item passes the term.
        """
        raise NotImplementedError


class ParentContractTerm[P: ParentT](ContractTerm[P], metaclass=ABCMeta):
    pass


class ChildContractTerm[I: ChildT, P: ParentT](ContractTerm[I], metaclass=ABCMeta):
    @abstractmethod
    def run(self, item: I, context: ContractContext, parent: P = None) -> bool:
        """
        Run this term on the given item and its parent.

        :param item: The item to check.
        :param context: The contract context to use.
        :param parent: The parent item that the given child `item` belongs to if available.
        :return: Boolean for if the item passes the term.
        """
        raise NotImplementedError


validator_type = Union[
    Callable[[ContractTerm, ItemT, ContractContext], bool],
    Callable[[ContractTerm, ItemT, ContractContext, ParentT | None], bool]
]


def validate_context(func: validator_type) -> validator_type:
    """Decorator to validate the context before running a term."""
    def wrapper(term: ContractTerm, item: ItemT, context: ContractContext, parent: ParentT = None) -> bool:
        """Validate the context before running the term."""
        if term.needs_manifest and (context.manifest is None or not isinstance(context.manifest, Manifest)):
            raise Exception(
                "This term requires a manifest to execute. Provide a manifest through the ContractContext."
            )
        if term.needs_catalog and (context.catalog is None or not isinstance(context.catalog, CatalogArtifact)):
            raise Exception(
                "This term requires a catalog to execute. Provide a catalog through the ContractContext."
            )

        if isinstance(term, ChildContractTerm):
            return func(term, item, context, parent)
        return func(term, item, context)

    return wrapper



from copy import deepcopy
from unittest import mock

import pytest
from dbt.artifacts.resources.v1.components import ColumnInfo
from dbt.contracts.graph.nodes import CompiledNode
from dbt_common.contracts.metadata import CatalogTable, ColumnMetadata
from faker import Faker

from dbt_contracts.contracts import ContractContext
from dbt_contracts.contracts.generators.column import ColumnPropertyGenerator, ColumnPropertiesGenerator, \
    SetColumnDescription, SetDataType
from tests.contracts.generators.test_core import ChildPropertiesGeneratorTester
from tests.contracts.generators.test_properties import SetDescriptionTester


class TestSetColumnDescription(SetDescriptionTester[ColumnInfo, ColumnMetadata]):
    @pytest.fixture
    def generator(self) -> SetColumnDescription:
        return SetColumnDescription()

    @pytest.fixture
    def item(self, node_column: ColumnInfo) -> ColumnInfo:
        return node_column

    def test_run(self, generator: SetDataType, item: ColumnInfo, catalog_column: ColumnMetadata, faker: Faker):
        item.description = faker.sentence()
        catalog_column.comment = faker.sentence()

        generator.overwrite = True

        assert generator.run(item, catalog_column)
        assert item.description == catalog_column.comment


class TestSetDataType:

    @pytest.fixture
    def generator(self) -> SetDataType:
        """Fixture for the property generator to test."""
        return SetDataType()

    def test_skips_on_empty_data_type(self, generator: SetDataType, node_column: ColumnInfo):
        original_data_type = node_column.data_type

        generator.overwrite = True

        assert not generator._set_data_type(node_column, data_type=None)
        assert not generator._set_data_type(node_column, data_type="")
        assert node_column.data_type == original_data_type

    def test_skips_on_not_overwrite(self, generator: SetDataType, node_column: ColumnInfo):
        original_data_type = "old data_type"
        node_column.data_type = original_data_type
        data_type = "new data_type"

        generator.overwrite = False

        assert not generator._set_data_type(node_column, data_type=data_type)
        assert node_column.data_type == original_data_type

    def test_skips_on_matching_data_type(self, generator: SetDataType, node_column: ColumnInfo):
        original_data_type = "int"
        node_column.data_type = original_data_type

        generator.overwrite = True

        assert not generator._set_data_type(node_column, data_type=original_data_type)
        assert node_column.data_type == original_data_type

    def test_valid_set(self, generator: SetDataType, node_column: ColumnInfo):
        original_data_type = "int"
        node_column.data_type = original_data_type
        data_type = "timestamp"

        generator.overwrite = True

        assert generator._set_data_type(node_column, data_type=data_type)
        assert node_column.data_type == data_type

    def test_run(self, generator: SetDataType, node_column: ColumnInfo, catalog_column: ColumnMetadata):
        node_column.data_type = "int"
        catalog_column.type = "timestamp"

        generator.overwrite = True

        assert generator.run(node_column, catalog_column)
        assert node_column.data_type == catalog_column.type


class TestColumnPropertiesGenerator(
    ChildPropertiesGeneratorTester[ColumnInfo, ColumnMetadata, ColumnPropertyGenerator]
):

    @pytest.fixture
    def generator(self) -> ColumnPropertiesGenerator:
        return ColumnPropertiesGenerator()

    @pytest.fixture
    def item(self, node_column: ColumnInfo) -> ColumnInfo:
        return node_column

    @pytest.fixture
    def parent(self, node: CompiledNode) -> CompiledNode:
        return node

    def test_merge_skips_on_no_table_in_database(
            self,
            generator: ColumnPropertiesGenerator,
            item: ColumnInfo,
            parent: CompiledNode,
            context: ContractContext,
    ):
        with (
            mock.patch("dbt_contracts.contracts.generators.column.get_matching_catalog_table", return_value=None),
            mock.patch.object(
                generator.__class__, "generators", new_callable=mock.PropertyMock, return_value=[]
            ) as mock_generators,
        ):
            assert not generator.merge(item, context=context, parent=parent)
            mock_generators.assert_not_called()

    def test_merge_skips_on_no_column_in_database(
            self,
            generator: ColumnPropertiesGenerator,
            item: ColumnInfo,
            parent: CompiledNode,
            context: ContractContext,
            catalog_table: CatalogTable,
    ):
        table = deepcopy(catalog_table)
        table.columns.pop(item.name)

        with (
            mock.patch("dbt_contracts.contracts.generators.column.get_matching_catalog_table", return_value=table),
            mock.patch.object(
                generator.__class__, "generators", new_callable=mock.PropertyMock, return_value=[]
            ) as mock_generators,
        ):
            assert not generator.merge(item, context=context, parent=parent)
            mock_generators.assert_not_called()

    def test_merge(
            self,
            generator: ColumnPropertiesGenerator,
            item: ColumnInfo,
            parent: CompiledNode,
            context: ContractContext,
            catalog_table: CatalogTable,
    ):
        with (
            mock.patch(
                "dbt_contracts.contracts.generators.column.get_matching_catalog_table",
                return_value=catalog_table
            ),
            mock.patch.object(
                generator.__class__, "generators", new_callable=mock.PropertyMock, return_value=[]
            ) as mock_generators,
            mock.patch.object(generator.description.__class__, "run", return_value=True) as mock_description,
            mock.patch.object(generator.data_type.__class__, "run", return_value=False) as mock_data_type,
        ):
            assert not generator.merge(item, context=context, parent=parent)

            mock_generators.return_value = [generator.description, generator.data_type]
            assert generator.merge(item, context=context, parent=parent)

            mock_description.assert_called_once_with(item, catalog_table.columns[item.name])
            mock_data_type.assert_called_once_with(item, catalog_table.columns[item.name])
